"""Evaluator for VivJson

- Evaluator: Constructor.
- Evaluator#evaluate: Evaluate the given statements.

Refer to:
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB site is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-30'

# pylint: disable=C0302

from copy import deepcopy
from functools import singledispatchmethod
from math import isinf, isnan
from .config import Config
from .error import Error
from .environment import Environment
from .statement import Literal, Identifier, Keyword, Blank, \
                       Array, Block, Binary, Parameter, \
                       Callee, CalleeRegistry, Call, Loop, \
                       Get, Set, Remove, \
                       Return, Injection, Value, LITERAL
from .standard import Standard
from .tokens import Token

class Evaluator:  # pylint: disable=R0902
    """Evaluator class

    Evaluate statement.

    Attributes:
        _config (Config or NoneType): Configuration
        _environment (Environment): Environment instance that is used
                                    to read/write variable.
        _binary_operations (dict): Token type -> method
        _assign_operations (dict): Token type -> method
        _stack (list[dict]): Stack of evaluated statement.
                             There are two purposes.
                               - Keep limit of the recursive called
                                 times.
                               - Avoid software runaway.
                             The each element is a dict that have
                             2 strings.
                               - "name": Block type or the class name
                                 of evaluated statement.
                               - "detail": The detail of evaluated
                                 statement.
        _max_array_size (int): Maximum array/block size
        _max_depth (int): Maximum recursive called times of evaluate
                          method
        _max_loop_times (int): Maximum loop times of "for", "while",
                               and so on.
        _REPLACED_LETTERS (dict): The escaped letter and real letter.
    """
    _REPLACED_LETTERS = {chr(0x08): 'b', chr(0x0C): 'f', chr(0x0A): 'n',
                         chr(0x0D): 'r', chr(0x09): 't',
                         '"': '"', '\\': '\\'}

    def __init__(self, config=None):
        """Initialize class.

        Args:
            config (Config or NoneType): Configuration if needed.
        """
        self._config = config
        self._environment = Environment()
        infinity = Config.get_infinity_alt(config)
        if isinstance(infinity, str):
            self._environment.set(infinity, float('inf'))
        nan = Config.get_nan_alt(config)
        if isinstance(nan, str):
            self._environment.set(nan, float('nan'))
        self._binary_operations = {
            Token.Type.PLUS: Evaluator._add,
            Token.Type.MINUS: Evaluator._subtract,
            Token.Type.MULTIPLY: Evaluator._multiply,
            Token.Type.DIVIDE: Evaluator._divide,
            Token.Type.MODULO: Evaluator._modulo,
            Token.Type.NOT: Evaluator._logical_invert,
            Token.Type.AND: Evaluator._logical_and,
            Token.Type.OR: Evaluator._logical_or,
            Token.Type.EQ: Evaluator._is_equal,
            Token.Type.NE: Evaluator._is_not_equal,
            Token.Type.LT: Evaluator._is_less_than,
            Token.Type.LE: Evaluator._is_less_than_or_equal,
            Token.Type.GT: Evaluator._is_greater_than,
            Token.Type.GE: Evaluator._is_greater_than_or_equal,
            Token.Type.IN: Evaluator._is_contained
        }
        self._assign_operations = {
            Token.Type.PLUS_ASSIGN: Evaluator._add,
            Token.Type.MINUS_ASSIGN: Evaluator._subtract,
            Token.Type.MULTIPLY_ASSIGN: Evaluator._multiply,
            Token.Type.DIVIDE_ASSIGN: Evaluator._divide,
            Token.Type.MODULO_ASSIGN: Evaluator._modulo
        }
        self._stack = []
        self._max_array_size = Config.get_max_array_size_alt(config)
        self._max_depth = Config.get_max_depth_alt(config)
        self._max_loop_times = Config.get_max_loop_times_alt(config)

    def get(self, name):
        """Get a variable's value.

        Args:
            name (str): Variable name

        Returns:
            Any: Its value
                 Environment.UNDEFINED is returned if the given name's
                 variable is not existed.

        Raises:
            EvaluateError: It is happen if the given name isn't string.
        """
        if not isinstance(name, str):
            self._abort(f'"{name}" is not string in Evaluator#get.')
        return self._environment.get(name, only_this_scope=True)

    def set(self, name, value):
        """Set a variable.

        Args:
            name (str): Variable name
            value (Any): Its value

        Raises:
            EvaluateError: It is happen if the given name isn't string.
        """
        if not isinstance(name, str):
            self._abort(f'"{name}" is not string in Evaluator#set.')
        self._environment.set(name, value, only_this_scope=True)

    def rewind_after_abort(self):
        """Rewinds stack and environment after abort.

        Fix as below. This is clean up for abort (exception).
        When Viv.run_method is failed, instance has unnecessary
        information. It may be happen unusual behavior in the next
        running via its instance. So clean up is needed.

        - Reset stack of evaluated statement "{@link #stack}".
        - Rewind environment "{@link #environment}" to first hierarchy.

        Note that it is assumed that the first call of evaluate method
        is evaluate(self, block: Block).

        For example,

          - The newest Environment #1  : The present position
              - enclosing = #2                |
          - Environment #2                    |
              - enclosing = #3                V
          - Environment #3             : Rewind to this position.
            (This is made by Block statement of making Instance.)
              - enclosing = #4
              - Instance's member variable is available.
              - Instance's method is available.
          - Root Environment #4
              - enclosing = null
        """
        # Reset stack of evaluated statement.
        self._stack = []
        # Rewind environment.
        while (enclosing := self._environment.get_enclosing()) is not None \
              and enclosing.get_enclosing() is not None:
            self._environment = enclosing

    @singledispatchmethod
    def evaluate(self, statement):
        """Evaluate statement.

        This should not be called. Sub-class, such as Identifier,
        should be called.

        Args:
            statement (Statement): The evaluated statement

        Raises:
            EvaluateError: This exception always happen when this
                           method is called.
        """
        self._abort(f'Cannot evaluate "{statement}"', statement)

    @evaluate.register
    def _(self, literal: Literal) -> LITERAL:
        """Evaluate literal.

        Args:
            literal (Literal): The evaluated literal statement

        Returns:
            float or int or str or bool or NoneType: the literal value
                                                as evaluated result.

        Raises:
            EvaluateError: It is happen if the given number is invalid.
        """
        self._enter_evaluate(literal)
        value = None
        if literal.token.type == Token.Type.NUMBER:
            try:
                value = float(literal.token.lexeme) \
                        if '.' in literal.token.lexeme \
                            or 'e' in literal.token.lexeme \
                            or 'E' in literal.token.lexeme \
                        else int(literal.token.lexeme)
            except (ValueError, OverflowError):
                self._abort(f'Invalid number "{literal}"', literal)
        elif literal.token.type == Token.Type.STRING:
            value = literal.token.lexeme
        elif literal.token.type == Token.Type.TRUE:
            value = True
        elif literal.token.type == Token.Type.FALSE:
            value = False
        self._leave_evaluate(literal)
        return value

    @evaluate.register
    def _(self, identifier: Identifier):
        """Evaluate identifier.

        Args:
            identifier (Identifier): The evaluated identifier statement

        Returns:
            Any: variable's value is returned mostly.
                 Identifier statement is returned if the given
                 identifier is Standard method name.
                 null if variable is undefined.

        Raises:
            EvaluateError: It is happen if the unexpected token is
                           included.
        """
        self._enter_evaluate(identifier)
        is_std_method = \
            Standard.get_method(identifier.name.lexeme) is not None

        if identifier.name.type != Token.Type.IDENTIFIER \
                and not is_std_method:
            self._abort(f'Cannot evaluate "{identifier}"', identifier)

        value = self._environment.get(None if identifier.name.lexeme == '.'
                                      else identifier.name.lexeme)
        if isinstance(value, Get):        # When it is alias,
            value = self.evaluate(value)  # read original variable's value.
        if value == Environment.UNDEFINED:
            if is_std_method:
                value = identifier
            else:
                value = None  # The undefined variable's value is null.
        self._leave_evaluate(identifier)
        return value

    @evaluate.register
    def _(self, keyword: Keyword):
        """Evaluate keyword.

        This should not be called.
        Keyword statement is used in Block/Loop statement. Because
        Keyword statement is "break" or "continue".
        So This is not used by evaluate method directly.

        Args:
            keyword (Keyword): The evaluated keyword statement

        Raises:
            EvaluateError: This exception always happen when this
                           method is called.
        """
        self._abort(f'Cannot evaluate "{keyword}"', keyword)

    @evaluate.register
    def _(self, _: Blank):
        """Evaluate blank.

        Args:
            _ (Blank): The evaluated blank statement

        Returns:
            NoneType: None is always returned.
        """
        return None

    @evaluate.register
    def _(self, array: Array) -> list:
        """Evaluate array.

        Args:
            array (Array): The evaluated array statement

        Returns:
            list: The evaluated result

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(array)
        if len(array.values) > self._max_array_size:
            self._abort('Array size is too large.', array)
        values = list(map(self.evaluate, array.values))
        self._leave_evaluate(array)
        return values

    @evaluate.register
    def _(self, block: Block) -> dict:
        """Evaluate block.

        Args:
            block (Block): The evaluated block statement

        Returns:
            Any: A dict as the evaluated block is returned mostly.
            However various value may be returned by ":=" and
            "return()".

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(block)

        # self._print_stack()

        previous_environment = self._environment
        self._environment = Environment(previous_environment)  # Add scope
        # When "return" statement is executed later, it should be recorded.
        # The certain variable is need in order to recording.
        # It is created at starting constructor or anonymous/pure function.
        is_recording = block.type in (Block.CLASS_CONSTRUCTOR,
                                Block.ANONYMOUS_FUNCTION, Block.PURE_FUNCTION)
        if is_recording:
            self._environment.set(Environment.FUNCTION_RETURN_NAME,
                                  Environment.UNDEFINED, only_this_scope=True)

        for value in block.values:
            reason = self._set_reason_going_back(value)
            if reason is not None:
                break

            self.evaluate(value)

            # Finish if break/continue/return happen in the above evaluation.
            reason = self._get_reason_going_back()
            if reason is not None:
                break

        result = Environment.UNDEFINED
        if is_recording:
            value = self._environment.get(Environment.FUNCTION_RETURN_NAME,
                                          only_this_scope=True)
            if value != Environment.FUNCTION_RETURN_NAME:
                result = value
        if result == Environment.UNDEFINED:
            result = self._environment.get()

        if block.type != Block.CLASS_CONSTRUCTOR:
            self._environment = previous_environment  # Restore scope
        self._leave_evaluate(block)
        return result

    @evaluate.register
    def _(self, binary: Binary):
        """Evaluate binary.

        For example, 3 + 2 is evaluated.
        Note that short-circuit evaluation (minimal evaluation) is happen
        for <and> or <or>.

        Args:
            binary (Binary): The evaluated binary statement

        Returns:
            Any: The result of binary operation if it is success, None
                 otherwise.

        Raises:
            EvaluateError: It is happen by some reason.
                           - The operator/operand is invalid.
                           - Array size reaches limit.
                           - Divide by 0.
        """
        self._enter_evaluate(binary)

        op_type = binary.operator.type
        if op_type not in self._binary_operations:
            self._abort(f'Cannot evaluate "{binary}"', binary)

        left = self.evaluate(binary.left)

        value = None
        # 1. Short-circuit evaluation (minimal evaluation) for <and>
        # 2. Short-circuit evaluation (minimal evaluation) for <or>
        # 3. Other evaluation
        if op_type == Token.Type.AND and not Evaluator.is_truthy(left):
            value = False
        elif op_type == Token.Type.OR and Evaluator.is_truthy(left):
            value = True
        else:
            method = self._binary_operations[op_type]
            right = self.evaluate(binary.right)
            try:
                value = method(left, right, self._config)
            except Error.EvaluateError:
                self._abort(f'Cannot evaluate "{binary}"', binary)

        self._leave_evaluate(binary)
        return value

    @evaluate.register
    def _(self, parameter: Parameter):
        """Evaluate parameter.

        This should not be called.
        Parameter statement is used in Call statement. Because
        Parameter statement is "class" or "function".
        So This is not used by evaluate method directly.

        Args:
            parameter (Parameter): the evaluated parameter statement

        Raises:
            EvaluateError: This exception always happen when this
                           method is called.
        """
        self._abort(f'Cannot evaluate "{parameter}"', parameter)

    @evaluate.register
    def _(self, callee: Callee):  # pylint: disable=R1711
        """Evaluate callee.

        Callee statement is the definition (entity) of function.
        For example, "function test() {print(10)}" is represented with
        Callee statement.
        This is wrapped with CalleeRegistry statement. And it is stored
        into Environment in order to register function.

        Args:
            callee (Callee): The evaluated callee statement

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(callee)
        self._environment.set(callee.name.name.name.lexeme,
                CalleeRegistry(callee, None, False))  # definition of function
        self._leave_evaluate(callee)
        return None

    @evaluate.register
    def _(self, callee_registry: CalleeRegistry):
        """Evaluate callee registry.

        This should not be called.
        CalleeRegistry statement is used to register function. This is
        not used by evaluate method directly.

        Args:
            callee_registry (CalleeRegistry): the evaluated 
                                              CalleeRegistry statement

        Raises:
            EvaluateError: This exception always happen when this
                           method is called.
        """
        self._abort(f'Cannot evaluate "{callee_registry}"', callee_registry)

    @evaluate.register
    def _(self, call: Call):
        """Evaluate call.

        Function call is done.

        Args:
            call (Call): The evaluated call statement

        Returns:
            Any: The returned value of function

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(call)

        if not isinstance(call.name, (Identifier, Get)):
            self._abort(f'Cannot call for {call}', call)

        result = None
        try:
            # User defined function
            result, is_done = self._call_function(call)
            if not is_done:
                # Standard library's method
                name = call.name
                value = self.evaluate(name)
                if isinstance(value, Identifier):
                    name = value.name.lexeme
                else:
                    if isinstance(name, Get):
                        if len(name.members) == 1:
                            name = name.members[0]
                    if isinstance(name, Identifier):
                        name = name.name.lexeme
                if not isinstance(name, str):
                    self._abort(f'Cannot call for {call}', call)
                method = Standard.get_method(name)
                if method is None:
                    self._abort(f'Cannot call for {call}', call)
                result = method(self, call, self._config)
        except Error.EvaluateError as e:
            message = str(e)
            is_baton = message != ''
            statement_for_location = None
            if not is_baton:
                message = f'{call} is failed'
                statement_for_location = call
            self._abort(message, statement_for_location,
                        eliminate_tag=is_baton)

        self._leave_evaluate(call)
        return result

    @evaluate.register
    def _(self, loop: Loop):  # pylint: disable=R0912
        """Evaluate loop.

        Args:
            loop (Loop): The evaluated loop statement

        Returns:
            Any: The returned value of Loop.
                 Loop has Block. Block is the function.
                 The function returns value.

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(loop)

        # Make sure that initial/continuous condition is given as list.
        if not isinstance(loop.initial, list) \
                or not isinstance(loop.continuous, list):
            self._abort(f'Loop condition of {loop.call} is invalid',
                        loop.call)
        # Make sure that correct iterator is given.
        iterator = None
        if loop.each is not None or loop.iterator is not None:
            if not isinstance(loop.each, Identifier) \
                    or not isinstance(loop.iterator, (list, dict)):
                self._abort(f'Iterator of {loop.call} is invalid', loop.call)
            # Deep copy is needed instead of Shallow copy.
            # Shallow copied value may be broken. For example, the value of
            # "list" is broken as below.
            # list=[{"x":3,"y":5}];for(a in list){a.x=2} --> [{"x":2, "y":5}]
            # list={"x":{"y":5}};for(a in list){a[1]["y"]=1} --> {"x":{"y":1}}
            iterator = deepcopy(loop.iterator)
            if isinstance(iterator, dict):
                iterator = list(iterator.items())

        # Add new environment
        previous_environment = self._environment
        self._environment = Environment(previous_environment)  # Add scope
        self._environment.set(Environment.BREAK_NAME,
                              Environment.UNDEFINED, only_this_scope=True)

        is_satisfied = True
        for loop_times in range(self._max_loop_times):
            statements = loop.initial if loop_times == 0 else loop.continuous
            for statement in statements:
                is_satisfied = self.evaluate(statement)
            is_satisfied = Evaluator.is_truthy(is_satisfied)
            if not is_satisfied:
                break

            self._environment.set(Environment.CONTINUE_NAME,
                                  Environment.UNDEFINED, only_this_scope=True)

            if iterator is not None:
                if len(iterator) == 0:
                    is_satisfied = False
                    break
                datum = iterator.pop(0)
                if isinstance(datum, tuple):
                    datum = list(datum)
                self._environment.set(loop.each.name.lexeme, datum,
                                      only_this_scope=True)

            reason = None
            for statement in loop.statements:
                reason = self._set_reason_going_back(statement)
                if reason is not None:
                    break

                self.evaluate(statement)

                # Terminate current loop if break/continue/return happen
                # in the above evaluation.
                reason = self._get_reason_going_back()
                if reason is not None:
                    break

            if reason in (Environment.BREAK_NAME,
                          Environment.FUNCTION_RETURN_NAME):
                is_satisfied = False
                break

        if is_satisfied:
            self._abort(f'Loop times reach maximum ({self._max_loop_times})',
                        loop.call)

        result = self._environment.get()

        # Discard current environment
        self._environment = previous_environment  # Restore scope

        self._leave_evaluate(loop)
        return result

    @evaluate.register
    def _(self, get: Get):  # pylint: disable=R0912
        """Evaluate get.

        This (Statement "Get") is familiar with Statement "Identifier".
        If variable is simple, such as foo, its value is given by
        evaluating Statement "Identifier" of foo.
        This (Statement "Get") is used to obtain the value of array and
        block, such as foo[2] and bar.baz.

        Args:
            get (Get): The evaluated get statement

        Returns:
            Any: variable's value

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(get)
        if len(get.members) == 0:
            self._abort(f'Cannot get {get}', get)
        variable = Environment.UNDEFINED
        for member in get.members:
            if variable is None:  # The undefined variable
                break
            member = self.evaluate(member)
            if variable == Environment.UNDEFINED:
                variable = member
                continue
            if isinstance(variable, (list, dict)):
                if isinstance(member, bool):  # Avoid boolean
                                    # Note that isinstance(True, int) is True.
                    self._abort(f'Cannot get {get}', get)
                if isinstance(member, float):
                    if member % 1 != 0:
                        self._abort(f'Cannot get {get}', get)
                    member = int(member)
                # for foo[bar] or foo.bar as array
                if isinstance(member, int) and isinstance(variable, list):
                    length = len(variable)
                    if member < 0:  # Backward access
                        member += length
                    if 0 <= member < length:
                        variable = variable[member]
                        continue
                    # null because out of range
                    variable = None
                    break
                # for foo[bar] of foo.bar as block
                if isinstance(variable, dict):
                    if isinstance(member, int):
                        member = str(member)
                    if isinstance(member, str) and member in variable:
                        variable = variable[member]
                        continue
                    # null because member is not found.
                    variable = None
                    break
            self._abort(f'Cannot get {get}', get)
        self._leave_evaluate(get)
        return None if variable == Environment.UNDEFINED else variable

    @evaluate.register
    def _(self, s: Set):  # pylint: disable=R0912
        """Evaluate set.

        For assignment, various situations are available as below.

        1. Simple assignment: x = 3
        2. Assignment for alias:
                function test(a) { a = 3, ... } --> _[0] = 3
        3. Assignment for block's member:
                a = {b: [2, 3]}, a.b[1] = 10 --> a = {b: [2, 10]}
        4. Removing: It is done with Remove statement.

        "Set" statement has parts of the above variable.
        For example, x["y"]["z"] is represented as ["x", "y", "z"].

        Variable may be alias.
        For example, x is alias of _[0] and x["y"]["z"] is given,
        _[0]["y"]["z"] is evaluated.

        Args:
            s (Set): The evaluated set statement

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(s)
        value = self.evaluate(s.value)
        if isinstance(value, (dict, list)):
            value = deepcopy(value)

        # The returned value of ":="
        if s.operator.type == Token.Type.RESULT:
            self._environment.set(None, value)  # None indicates
                                                # the returned value.
            self._leave_evaluate(s)
            return None  # None is meaningless.

        is_local = s.operator.type == Token.Type.SET  # ":" makes local.

        if len(s.members) == 0:
            self._abort(f'Cannot assign {s}', s)

        # Get the head. For example, "x" of x["y"]["z"]
        member = s.members[0]
        if not isinstance(member, Identifier):
            member = self.evaluate(member)
            if not isinstance(member, Identifier):
                self._abort(f'Cannot assign {s}', s)

        # Avoid system variable.
        name = member.name.lexeme
        if len(name) >= 6 and name[0:3] == '___' and name[-3:] == '___':
            self._abort(f'Cannot assign {s} '
                        "because variable's prefix/suffix is '___'", s)

        members = []
        maybe_alias = self._environment.get(name)
        if isinstance(maybe_alias, Get):  # for alias
            # For example, maybe_alias = ["_", 1]
            #              "_" is Identifier and 1 is Literal
            if len(maybe_alias.members) <= 1 \
                    or not isinstance(maybe_alias.members[0], Identifier):
                self._abort(f'Cannot assign {s}', s)
            name = maybe_alias.members[0].name.lexeme  # It is '_'.
            # Get the following member. For example, ["y"]["z"] of x["y"]["z"]
            members = s.members.copy()
            members.pop(0)
            # Conjoin.
            # For example, ["_", 1, "y", "z"]
            for index, member in enumerate(maybe_alias.members):
                members.insert(index, member)
        else:  # except alias
            if len(s.members) == 1:  # for simple variable, such as i, num.
                if s.operator.type in self._assign_operations:  # +=, -=, ...
                    method = self._assign_operations[s.operator.type]
                    original_value = self._environment.get(name, is_local)
                    try:
                        value = method(original_value, value, self._config)
                    except Error.EvaluateError:
                        self._abort(f'Cannot evaluate "{s}"', s)
                if s.operator.type == Token.Type.REMOVE:
                    self._environment.remove(name)
                else:
                    if isinstance(value, CalleeRegistry):
                        value = CalleeRegistry(value.callee,
                                               value.environment,
                                               True)  # Reference
                    self._environment.set(name, value, is_local)
                self._leave_evaluate(s)
                return None  # None is meaningless.
            members = s.members.copy()
        members[0] = Literal(Token(Token.Type.STRING, lexeme=name))
        variable = {name: self._environment.get(name, is_local)}
        self._set_member(variable, members, value, s.operator.type)
        self._environment.set(name, variable[name], is_local)

        self._leave_evaluate(s)
        return None  # None is meaningless.

    @evaluate.register
    def _(self, remove: Remove):  # pylint: disable=R1711
        """Evaluate remove.

        Args:
            remove (Remove): The evaluated remove statement

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(remove)

        s = Set(remove.members, remove.token, Literal(Token(Token.Type.NULL)))
        try:
            self.evaluate(s)
        except Error.EvaluateError as e:
            message = str(e)
            is_baton = message != ''
            statement_for_location = None
            if not is_baton:
                message = f'{remove} is failed'
                statement_for_location = remove
            else:
                message = message.replace('assign', 'remove')
            self._abort(message, statement_for_location,
                        eliminate_tag=is_baton)

        self._leave_evaluate(remove)
        return None  # None is meaningless.

    @evaluate.register
    def _(self, r: Return):
        """Evaluate return.

        This should not be called.
        Return statement is used in Block/Loop statement. This is
        not used by evaluate method directly.

        Args:
            r (Return): The evaluated Return statement

        Raises:
            EvaluateError: This exception always happen when this
                           method is called.
        """
        self._abort(f'Cannot evaluate "{r}"', r)

    @evaluate.register
    def _(self, injection: Injection):  # pylint: disable=R1711
        """Evaluate injection.

        The following values are accepted.
        - null
        - Boolean
        - Number (Byte, Short, Integer, Long, Float, Double)
        - String
        - ArrayList<@Nullable Object>
        - HashMap<String, @Nullable Object>

        Args:
            injection (Injection): The evaluated injection statement

        Raises:
            EvaluateError: It is happen if the prefix/suffix of
                           variable name is "___" or value is
                           illegal.
        """
        self._enter_evaluate(injection)
        if injection.value is not None and not isinstance(injection.value,
                                        (bool, int, float, str, list, dict)):
            self._abort(f'Cannot use a value of "{injection}"', injection)
        if len(injection.variable) >= 6 and injection.variable[0:3] == '___' \
                and injection.variable[-3:] == '___':
            self._abort(f'Cannot use a value of "{injection}" because '
                        "variable's prefix/suffix is '___'", injection)
        self._environment.set(injection.variable, injection.value)
        self._leave_evaluate(injection)
        return None  # None is meaningless.

    @evaluate.register
    def _(self, value: Value):
        """Evaluate value.

        The following values are accepted.
        - null
        - Boolean
        - Number (Byte, Short, Integer, Long, Float, Double)
        - String
        - ArrayList<@Nullable Object>
        - HashMap<String, @Nullable Object>

        Args:
            value (Value): The evaluated value statement

        Raises:
            EvaluateError: It is happen if value is illegal.
        """
        self._enter_evaluate(Value)
        if value.value is not None and not isinstance(value.value,
                                        (bool, int, float, str, list, dict)):
            self._abort(f'Cannot use a value of "{value}"', value)
        self._leave_evaluate(Value)
        return value.value

    def _enter_evaluate(self, statement):
        """Record to enter new evaluation.

        Args:
            statement (Statement or NoneType): Data class of new
                                               evaluation

        Raises:
            EvaluateError: It is happen if the recursive call times
                           reach limit.
        """
        if len(self._stack) >= self._max_depth:
            self._abort('Recursive call times reach '
                        f'maximum ({self._max_depth}).')
        evaluated = {
            'name': statement.type if isinstance(statement, Block) \
                    else statement.__class__.__name__,
            'detail': str(statement)
        }
        self._stack.append(evaluated)

    def _leave_evaluate(self, statement):
        """Record to leave current evaluation.

        Args:
            statement (Statement or NoneType): Data class of current
                                               evaluation

        Raises:
            EvaluateError: It is happen if the order of
                           evaluating/leaving is invalid
                           or recorded data is invalid.
        """
        evaluated = None
        try:
            evaluated = self._stack.pop()
        except IndexError:
            pass
        name = statement.type if isinstance(statement, Block) \
               else statement.__class__.__name__
        if not isinstance(evaluated, dict) or "name" not in evaluated \
                or evaluated['name'] != name:
            self._abort('Interpreter side error in Evaluator#_leave_evaluate')

    def _print_stack(self):
        """Print stack

        Raises:
            EvaluateError: it is happen if recorded data is invalid.
        """
        for index, evaluated in enumerate(self._stack):
            if not isinstance(evaluated, dict) or "name" not in evaluated \
                    or "detail" not in evaluated:
                self._abort('Interpreter side error '
                            'in Evaluator#_print_stack')
            print(f'> {index}: {evaluated["name"]} ({evaluated["detail"]})')

    def _get_reason_going_back(self):
        """Get reason of going back until function/loop statement.

        Returns:
            str or NoneType: Reason if going back is needed,
                             None otherwise.
                             Reason is which one:
                             - Environment.BREAK_NAME
                             - Environment.CONTINUE_NAME
                             - Environment.FUNCTION_RETURN_NAME
        """
        for name in (Environment.FUNCTION_RETURN_NAME,
                     Environment.BREAK_NAME, Environment.CONTINUE_NAME):
            value = self._environment.get(name)
            if value != Environment.UNDEFINED:
                return name
        return None

    def _set_reason_going_back(self, statement):
        """Set reason of going back until function/loop statement.

        Args:
            statement: Statement that may be return/break/continue.

        Returns:
            str or NoneType: Reason if going back is needed,
                             None otherwise.
                             Reason is which one:
                             - Environment.BREAK_NAME
                             - Environment.CONTINUE_NAME
                             - Environment.FUNCTION_RETURN_NAME

        Raises:
            EvaluateError: It is happen if the returned value is
                           invalid or break/continue is used without
                           loop.
        """
        # for "return"
        if isinstance(statement, Return):
            reason = Environment.FUNCTION_RETURN_NAME
            value = reason
            if statement.value is not None:
                # When function object is returned, the persistent
                # environment is given for closure.
                value = self.evaluate(statement.value)
                if isinstance(value, CalleeRegistry):
                    value = CalleeRegistry(value.callee,
                                           self._environment,  # Closure
                                           True)  # Reference
            self._environment.set(reason, value)
            return reason

        # for "break"/"continue"
        reason = None
        if isinstance(statement, Keyword):
            if statement.token.type == Token.Type.BREAK:
                reason = Environment.BREAK_NAME
            elif statement.token.type == Token.Type.CONTINUE:
                reason = Environment.CONTINUE_NAME
        if reason is None:
            return None
        # "break"/"continue" is allowed for Loop.
        # So we find Loop. However function is found, finding is terminated.
        found = False
        for evaluated in reversed(self._stack):
            if not isinstance(evaluated, dict) or 'name' not in evaluated:
                break
            if evaluated['name'] == 'Loop':
                found = True
                break
            if evaluated['name'] in (Block.CLASS_CONSTRUCTOR,
                            Block.ANONYMOUS_FUNCTION, Block.PURE_FUNCTION):
                break
        if not found:
            self._abort(f'{statement.token.lexeme} is found without loop',
                        statement)
        self._environment.set(reason, True)
        return reason

    def _call_function(self, call: Call):  # pylint: disable=R0912, R0914
        """Call a function.

        Args:
            call (Call): The information that contains function name,
                         arguments, and so on.

        Returns:
            Any, bool: 2 values are returned.
                       1st value: The result of function.
                       2nd value: True if processing is done.

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        # Get function entity.
        statement = call.name
        if isinstance(statement, Identifier):
            statement = self._environment.get(statement.name.lexeme)
        if isinstance(statement, Get):  # for alias of argument or
                                        # variable that is assigned function
            statement = self.evaluate(statement)
        if isinstance(statement, Block):  # for anonymous function
            return self.evaluate(statement), True
        if not isinstance(statement, CalleeRegistry):
            return None, False
        callee = statement.callee
        environment = statement.environment

        # Check its modifier.
        modifier = callee.name.modifier
        if modifier is None \
                or modifier.type != Token.Type.FUNCTION:
            return None, False

        # Get parameter's definition and function's statements of Callee.
        # They are contained within an Array.
        # For example, there are 2 parameters: This array is
        # [1st parameter, 2nd parameter, A block of statements].
        parameter_length = len(callee.parameters.values)
        if parameter_length < 1:  # One parameter of block is needed at least
            raise Error.EvaluateError()
        parameter_length -= 1  # Number of actual parameters except block
        statements = callee.parameters.values[-1]  # Block
        if not isinstance(statements, Block):
            raise Error.EvaluateError()

        # Add scope for pure function.
        # Reuse environment for closure.
        previous_environment = self._environment
        self._environment = environment if environment is not None \
                            else Environment(previous_environment)

        # Make implicit variable from arguments.
        implicit = []
        for index, value in enumerate(call.arguments.values):
            is_function = False
            is_reference = False
            if index < len(callee.parameters.values):
                parameter = callee.parameters.values[index]
                if isinstance(parameter, Parameter) \
                        and isinstance(parameter.modifier, Token):
                    is_function = \
                        parameter.modifier.type == Token.Type.FUNCTION
                    is_reference = \
                        parameter.modifier.type == Token.Type.REFERENCE
            if not is_function:
                value = self.evaluate(value)
            if not is_reference and isinstance(value, (list, dict)):
                value = deepcopy(value)
            elif isinstance(value, tuple):  # just in case
                value = list(value)
            implicit.append(value)
        self._environment.set('_', implicit, only_this_scope=True)

        # Set parameter as alias of implicit variable.
        for i in range(parameter_length):
            if not isinstance(callee.parameters.values[i], Parameter):
                raise Error.EvaluateError()
            name = callee.parameters.values[i].name.name  # Token
            if name.type != Token.Type.IDENTIFIER:
                raise Error.EvaluateError()
            value = Get([
                Identifier(Token(Token.Type.IDENTIFIER, lexeme='_')),
                Literal(Token(Token.Type.NUMBER, lexeme=str(i)))
            ]) if i < len(implicit) \
            else None
            self._environment.set(name.lexeme, value, only_this_scope=True)

        # Evaluate
        result = self.evaluate(statements)

        self._environment = previous_environment  # Restore scope

        return result, True

    def _set_member(self, target, members, value, operator
                   ):  # pylint: disable=R0912, R0915
        """Set/Modify/Remove a member within variable.

        For example, a.c.1 = 20 is set in
                     {"a": {"b": true, "c": [false, 100]}} .

        1. Caller has members and value.
           - Members: ["a", "c", 1]
           - Value: 20

        2. Caller gets original whole value of "a" from Environment.
           - Original whole value: {"b": true, "c": [false, 100]}

        3. Caller makes block that corresponds to members ["a","c",1].
           That is, it is made that key-value pairs of "a" and its
           value.
           - Block: {"a": {"b": true, "c": [false, 100]}}

        4. Caller call this method.
           - target: {"a": {"b": true, "c": [false, 100]}}

           This method is called as below.
           1st call is done by the above 4th step.
           2nd-4th call is done itself recursively.
           | #   | target                                | members       |
           |-----|---------------------------------------|---------------|
           | 1st | {"a": {"b": true, "c": [false, 100]}} | ["a", "c", 1] |
           | 2nd | {"b": true, "c": [false, 100]}        | ["c", 1]      |
           | 3rd | [false, 100]                          | [1]           |
           | 4th | 100                                   | []            |

           | #   | Returned value                       |
           |-----|--------------------------------------|
           | 4th | 20                                   |
           | 3rd | [false, 20]                          |
           | 2nd | {"b": true, "c": [false, 20]}        |
           | 1st | {"a": {"b": true, "c": [false, 20]}} |

        5. Caller gets the new value of "a" from the returned value:
           {"b": true, "c": [false, 20]}

        6. Caller put it into Environment.

        Args:
            target (Any): Target variable or value.
                          When this method is called from outside,
                          it must be dict.
                          When this method is called from itself
                          recursively, it is dict or list except
                          last call.
                          It must be a value of the latest member
                          at last call.
            members (list[Statement]): Members of the above target
            value (Any): The latest member's value
            operator (Token.Type) Operator's token type

        Returns:
            Any: Variable or Value

        Raises:
            EvaluateError: It is happen if evaluation is failed.
        """
        self._enter_evaluate(None)

        if not isinstance(target, (dict, list)) or len(members) == 0:
            if len(members) > 0:
                self._abort(f'Cannot assign for "{members[0]}"', members[0])
            if operator in self._assign_operations:
                method = self._assign_operations[operator]
                # The following method (_add, _subtract, and so on) may be
                # raise EvaluateError without error-message.
                # However, in present, error position is unknown.
                # This is resolved in the previous recursive caller.
                value = method(target, value, self._config)
            if operator == Token.Type.REMOVE:
                value = Environment.UNDEFINED  # #1
            if isinstance(value, CalleeRegistry):
                value = CalleeRegistry(value.callee,
                                       value.environment,
                                       True)  # Reference
            self._leave_evaluate(None)
            return value

        member = members.pop(0)
        evaluated_member = self.evaluate(member)

        if isinstance(evaluated_member, bool):  # Avoid boolean
                                    # Note that isinstance(True, int) is True.
            self._abort(f'Boolean "{member}" is not suitable', member)

        if isinstance(evaluated_member, float):
            if evaluated_member % 1 != 0:
                self._abort(f'Floating-point number "{member}" is invalid',
                            member)
            evaluated_member = int(evaluated_member)
        if isinstance(evaluated_member, int) and isinstance(target, list):
            length = len(target)
            if evaluated_member < 0:  # Backward access
                evaluated_member += length
            if evaluated_member < 0 or evaluated_member >= length:
                self._abort(f'Index "{evaluated_member}" is out of range',
                            member)
            try:
                value = self._set_member(target[evaluated_member],
                                        members, value, operator)
            except Error.EvaluateError:
                self._abort(f'Cannot evaluate "{member}"', member)
            if value == Environment.UNDEFINED:  # #1
                del target[evaluated_member]
            else:
                target[evaluated_member] = value
            self._leave_evaluate(None)
            return target

        if not isinstance(target, dict):
            self._abort(f'Invalid access for {member}', member)
        if isinstance(evaluated_member, int):
            evaluated_member = str(evaluated_member)
        if not isinstance(evaluated_member, str):
            self._abort(f'Cannot assign for "{member}"', member)
        if evaluated_member not in target:
            if len(members) > 0:
                self._abort(f'Cannot assign for "{members[0]}"',
                            members[0])
            if operator != Token.Type.REMOVE:
                target[evaluated_member] = value
        else:
            try:
                value = self._set_member(target[evaluated_member],
                                        members, value, operator)
            except Error.EvaluateError:
                self._abort(f'Cannot evaluate "{member}"', member)
            if value == Environment.UNDEFINED:  # #1
                del target[evaluated_member]
            else:
                target[evaluated_member] = value
        self._leave_evaluate(None)
        return target

    def _abort(self, message, statement=None, eliminate_tag=False):
        """Abort Evaluator.

        Print an error message and raise Exception.

        Args:
            message (str): Error message
            statement (Statement or NoneType, optional): Statement
                    where error occurs. It is used to get location.
                    None is given when cause is not Statement.
            eliminate_tag (bool, optional): True if tag is not needed,
                                            False otherwise.
                                            The default is False

        Raises:
            EvaluateError: This exception always happen when this method
                           is called.
        """
        raise Error.make_evaluate_error(message, statement, self._config,
                                        eliminate_tag)

    @staticmethod
    def _add(left, right, config=None):  # pylint: disable=R0911, R0912
        """Add

        | left \\ right | block | array | string | int     | float   | boolean | null    |
        |---------------|-------|-------|--------|---------|---------|---------|---------|
        | block         | block | array | ERROR  | ERROR   | ERROR   | ERROR   | block   |
        | array         | array | array | array  | array   | array   | array   | array   |
        | string        | ERROR | array | string | string  | string  | string  | string  |
        | int           | ERROR | array | string | int     | float   | boolean | int     |
        | float         | ERROR | array | string | float   | float   | boolean | float   |
        | boolean       | ERROR | array | string | boolean | boolean | boolean | boolean |
        | null          | block | array | string | int     | float   | boolean | null    |

        For example, {"a": 3, "b": 100} + {"0": "x"}
                                       --> {"a": 3, "b": 100, "0": "x"}
                     {"a": 3, "b": 100} + {"a": -5}
                                       --> {"a": -2, "b": 100}
                     {"a": 3} + [3, "a"] --> [{"a": 3}, 3, "a"]
                     [3, "a"] + {"a": 3} --> [3, "a", {"a": 3}]
                     {"a": 3} + "a" --> ERROR
                     [3, "a"] + true --> [3, "a", true]
                     true + [3, "a"] --> [true, 3, "a"]
                     [3, "a"] + [true, false] --> [3, "a", [true, false]]
                     3 + null --> 3
                     3 + "a" --> "3a"
                     3 + false --> true  # 3 is equivalent to true.
                                 # Then true + false --> true or false.
                     3 + 2 --> 5

        Raises:
            EvaluateError: It is happen if the operator/operand is
                           invalid or array/block size reaches limit.
        """
        max_array_size = Config.get_max_array_size_alt(config)
        # Result is array.
        is_left_array = isinstance(left, list)
        is_right_array = isinstance(right, list)
        if is_left_array:
            if len(left) + 1 > max_array_size:
                raise Error.EvaluateError()
            new_array = left.copy()
            new_array.append(right)
            return new_array
        if is_right_array:
            if 1 + len(right) > max_array_size:
                raise Error.EvaluateError()
            new_array = right.copy()
            new_array.insert(0, left)
            return new_array
        # Result is another operand.
        if left is None:
            return right
        if right is None:
            return left
        # Result is block.
        is_left_block = isinstance(left, dict)
        is_right_block = isinstance(right, dict)
        if is_left_block and is_right_block:
            length = len(left)
            new_block = left.copy()
            for right_name, right_value in right.items():
                if right_name in new_block:
                    new_block[right_name] = \
                            Evaluator._add(new_block[right_name], right_value,
                                           config)
                else:
                    length += 1
                    if length > max_array_size:
                        raise Error.EvaluateError()
                    new_block[right_name] = right_value
            return new_block
        if is_left_block or is_right_block:
            raise Error.EvaluateError()
        # Result is string.
        if isinstance(left, str):
            return f'{left}{Evaluator.make_str(right, config=config)}'
        if isinstance(right, str):
            return f'{Evaluator.make_str(left, config=config)}{right}'
        # Result is boolean.
        if isinstance(left, bool) or isinstance(right, bool):
            return Evaluator.is_truthy(left) or Evaluator.is_truthy(right)
        # Result is int or float.
        if Evaluator._is_numbers(left, right):
            return left + right
        raise Error.EvaluateError()

    @staticmethod
    def _subtract(left, right,  # pylint: disable=R0911, R0912, R0914
                  config=None):
        """Subtract

        | left \\ right | block | array  | string | int   | float | boolean | null    |
        |---------------|-------|--------|--------|-------|-------|---------|---------|
        | block         | block | #      | block  | ERROR | ERROR | ERROR   | block   |
        | array         | array | array  | array  | array | array | array   | array   |
        | string        | ERROR | *      | string | ERROR | ERROR | ERROR   | string  |
        | int           | ERROR | ERROR  | ERROR  | int   | float | ERROR   | int     |
        | float         | ERROR | ERROR  | ERROR  | float | float | ERROR   | float   |
        | boolean       | ERROR | ERROR  | ERROR  | ERROR | ERROR | ERROR   | boolean |
        | null          | ERROR | ERROR  | ERROR  | ERROR | ERROR | ERROR   | null    |

        # is block or ERROR
        * is string or ERROR

        For example, {"a": 10, "b": 20, "c": 30} - {"b": 5, "c": 10}
                                    --> {"a": 10, "b": 15, "c": 20}
                     {"a": 10, "b": 20} - {"b": 5, "c": 10}
                                    --> {"a": 10, "b": 15, "c": -10}
                     {"a": 10, "b": 20, "c": 30} - ["b", "c"]
                                    --> {"a": 10}
                     {"a": 10, "b": 20, "c": 30} - ["b", "c", 3]
                                    --> ERROR because there is number in array
                     {"a": 10, "b": 20, "c": 30} - "b"
                                    --> {"a": 10, "c": 30}
                     {"a": 10, "b": 20, "c": 30} - "d"
                                    --> {"a": 10, "b": 20, "c": 30}
                     [3, "a", "a"] - "a" --> [3]
                     [3, [100, true]] - [100, true] --> [3]
                     [3, {"a": null}] - {"a": null} --> [3]
                     [3, false, "a"] - true --> [3, false, "a"]
                     "large-dog&small-dog&2cat" - "dog"
                                    --> "large-&small-&2cat"
                     "large-dog&small-dog&2cat" - ["large-", "small-"]
                                    --> "dog&dog&2cat"
                     "large-dog&small-dog&2cat" - 2
                                    --> ERROR
                     "large-dog&small-dog&2cat" - ["large-", 2]
                                    --> ERROR
                     {"a": null} - null --> {"a": null}
                     100 - null --> 100
                     true - null --> true
                     null - "a" --> ERROR

        Raises:
            EvaluateError: It is happen if the operator/operand is
                           invalid or array/block size reaches limit.
        """
        if left is None and right is not None:
            raise Error.EvaluateError()
        # Result is array.
        is_left_array = isinstance(left, list)
        is_right_array = isinstance(right, list)
        if is_left_array:
            new_array = left.copy()
            Evaluator._remove_element_from_array(new_array, right)
            return new_array
        # Result is another operand.
        if right is None:
            return left
        # Result is block.
        is_left_block = isinstance(left, dict)
        is_right_block = isinstance(right, dict)
        if is_left_block:
            if is_right_block:
                max_array_size = Config.get_max_array_size_alt(config)
                new_block = left.copy()
                length = len(new_block)
                for right_name, right_value in right.items():
                    if right_name in new_block:
                        new_block[right_name] = \
                                Evaluator._subtract(new_block[right_name],
                                                    right_value, config)
                    else:
                        length += 1
                        if length > max_array_size:
                            raise Error.EvaluateError()
                        new_block[right_name] = \
                                Evaluator._multiply(-1, right_value, config)
                return new_block
            if is_right_array:
                new_block = left.copy()
                for name in right:
                    if not isinstance(name, str):
                        raise Error.EvaluateError()
                    if name in new_block:
                        del new_block[name]
                return new_block
            if isinstance(right, str):
                if right not in left:
                    return left
                new_block = left.copy()
                del new_block[right]
                return new_block
        elif is_right_block:
            raise Error.EvaluateError()
        # Result is string.
        if isinstance(left, str):
            elements = right if is_right_array else [right]
            for element in elements:
                if not isinstance(element, str):
                    raise Error.EvaluateError()
                left = left.replace(element, '')
            return left
        # Result is int or float.
        if Evaluator._is_numbers(left, right):
            return left - right
        raise Error.EvaluateError()

    @staticmethod
    def _multiply(left, right, config=None):  # pylint: disable=R0911, R0912
        """Multiply

        | left \\ right | block | array  | string | int    | float  | boolean | null |
        |---------------|-------|--------|--------|--------|--------|---------|------|
        | block         | block | ERROR  | ERROR  | ERROR  | ERROR  | ERROR   | null |
        | array         | ERROR | ERROR  | string | array  | array  | ERROR   | null |
        | string        | ERROR | string | ERROR  | string | string | ERROR   | null |
        | int           | ERROR | array  | string | int    | float  | ERROR   | null |
        | float         | ERROR | array  | string | float  | float  | ERROR   | null |
        | boolean       | ERROR | ERROR  | ERROR  | ERROR  | ERROR  | ERROR   | null |
        | null          | null  | null   | null   | null   | null   | null    | null |

        For example, {"a": 2} * {"a": 10, "b": 3} --> {"a": 20, "b": null}
                     {"a": 2} * {"b": 3} --> {"a": null, "b": null}
                     [3, "a"] * 2 --> [3, "a", 3, "a"]
                     [3, "a"] * "|" --> '3|a'
                     [3, "a"] * true -> ERROR
                     false * [3, "a"] --> ERROR
                     "3a" * 2  --> "3a3a"
                     100 * null --> null
                     [3, "a"] * [1, 2] --> ERROR

        Raises:
            EvaluateError: It is happen if the operator/operand is
                           invalid or array/block size reaches limit.
        """
        # Result is null.
        if left is None or right is None:
            return None
        # For boolean
        if isinstance(left, bool) or isinstance(right, bool):
            raise Error.EvaluateError()
        # Result is block.
        max_array_size = Config.get_max_array_size_alt(config)
        if isinstance(left, dict) and isinstance(right, dict):
            new_block = {}
            either = []
            for name in left.keys():
                if name not in right:
                    either.append(name)
                    continue
                new_block[name] = \
                        Evaluator._multiply(left[name], right[name], config)
            for name in right.keys():
                if name not in left:
                    either.append(name)
            if len(new_block) + len(either) > max_array_size:
                raise Error.EvaluateError()
            for name in either:
                new_block[name] = None
            return new_block
        # Result is string x n.
        if isinstance(left, str) and Evaluator._is_number(right):
            return left * int(right)
        if isinstance(right, str) and Evaluator._is_number(left):
            return right * int(left)
        # Result is array (elements x n).
        if isinstance(left, list) and Evaluator._is_number(right) \
                and len(left) * int(right) <= max_array_size:
            return left * int(right)
        if isinstance(right, list) and Evaluator._is_number(left) \
                and len(right) * int(left) <= max_array_size:
            return right * int(left)
        # Result is string.
        if isinstance(left, list) and isinstance(right, str):
            return right.join(Evaluator._make_str_array(left, config=config))
        if isinstance(left, str) and isinstance(right, list):
            return left.join(Evaluator._make_str_array(right, config=config))
        # Result is int/float.
        if Evaluator._is_numbers(left, right):
            return left * right
        raise Error.EvaluateError()

    @staticmethod
    def _divide(left, right, config=None):  # pylint: disable=R0912
        """Divide

        | left \\ right | block | array | string | int   | float | boolean | null  |
        |---------------|-------|-------|--------|-------|-------|---------|-------|
        | block         | #     | ERROR | ERROR  | ERROR | ERROR | ERROR   | ERROR |
        | array         | ERROR | ERROR | ERROR  | ERROR | ERROR | ERROR   | ERROR |
        | string        | ERROR | ERROR | array  | ERROR | ERROR | ERROR   | ERROR |
        | int           | ERROR | ERROR | ERROR  | *     | *     | ERROR   | ERROR |
        | float         | ERROR | ERROR | ERROR  | *     | *     | ERROR   | ERROR |
        | boolean       | ERROR | ERROR | ERROR  | ERROR | ERROR | ERROR   | ERROR |
        | null          | null  | null  | null   | x     | x     | null    | ERROR |

        # is block or ERROR
        * is int or float or ERROR
        x is null or ERROR

        For example, {"a": 2} / {"a": 10, "b": 3} --> {"a": 0.2, "b": null}
                     {"a": 2} / {"b": 3} --> ERROR
                     "a,b,c" / "," --> ["a", "b", "c"]
                     "a,b,c," / "," --> ["a", "b", "c", ""]
                     "a,b,c" / "" --> ["a", ",", "b", ",", "c"]
                     3 / 2 --> 1.5
                     3 / 0 --> ERROR
                     3 / null --> ERROR
                     null / 3 --> null
                     null / 0 --> ERROR

        Raises:
            EvaluateError: It is happen if the operator/operand is
                           invalid or array/block size reaches limit.
        """
        # Result is ERROR.
        if right is None \
                or (Evaluator._is_number(right) and right == 0):
            raise Error.EvaluateError()
        # Result is null.
        if left is None:
            return None
        # Result is block.
        max_array_size = Config.get_max_array_size_alt(config)
        if isinstance(left, dict) and isinstance(right, dict):
            new_block = {}
            for name in left.keys():
                if name not in right:
                    raise Error.EvaluateError()
                new_block[name] = \
                        Evaluator._divide(left[name], right[name], config)
            length = len(new_block)
            for name in right.keys():
                if name not in left:
                    length += 1
                    if length > max_array_size:
                        raise Error.EvaluateError()
                    new_block[name] = None
            return new_block
        # Result is array of string.
        if isinstance(left, str) and isinstance(right, str):
            if right == "":
                if len(left) > max_array_size:
                    raise Error.EvaluateError()
                array = []
                for letter in left:
                    array.append(letter)
            else:
                delimiter_length = len(right)
                count = 1
                index = -delimiter_length
                while index < len(left):
                    try:
                        index = left.index(right, index + delimiter_length)
                    except ValueError:
                        break
                    count += 1
                if count > max_array_size:
                    raise Error.EvaluateError()
                array = left.split(right)
            return array
        # Result is int or float.
        if Evaluator._is_numbers(left, right):
            result = left / right
            if result % 1 == 0:
                result = int(result)
            return result
        raise Error.EvaluateError()

    @staticmethod
    def _modulo(left, right, config=None):
        """Modulo

        | left \\ right | block | array | string | int  | float | boolean | null  |
        |---------------|-------|-------|--------|------|-------|---------|-------|
        | block         | #     | ERROR | ERROR  | ERROR| ERROR | ERROR   | ERROR |
        | array         | ERROR | ERROR | ERROR  | ERROR| ERROR | ERROR   | ERROR |
        | string        | ERROR | ERROR | ERROR  | ERROR| ERROR | ERROR   | ERROR |
        | int           | ERROR | ERROR | ERROR  | *    | *     | ERROR   | ERROR |
        | float         | ERROR | ERROR | ERROR  | *    | *     | ERROR   | ERROR |
        | boolean       | ERROR | ERROR | ERROR  | ERROR| ERROR | ERROR   | ERROR |
        | null          | null  | null  | null   | x    | x     | null    | ERROR |

        # is block or ERROR
        * is int or float or ERROR
        x is null or ERROR

        For example, {"a": 20} % {"a": 6, "b": 3} --> {"a": 2, "b": null}
                     {"a": 2} % {"b": 3} --> ERROR
                     15 % 6 --> 3
                     15 % 0 --> ERROR
                     null % 2 --> null
                     null / 0 --> ERROR
                     5 % null --> ERROR

        Raises:
            EvaluateError: It is happen if the operator/operand is
                           invalid.
        """
        # Result is ERROR.
        if right is None \
                or (Evaluator._is_number(right) and right == 0):
            raise Error.EvaluateError()
        # Result is null.
        if left is None:
            return None
        # Result is block.
        max_array_size = Config.get_max_array_size_alt(config)
        if isinstance(left, dict) and isinstance(right, dict):
            new_block = {}
            for name in left.keys():
                if name not in right:
                    raise Error.EvaluateError()
                new_block[name] = \
                        Evaluator._modulo(left[name], right[name], config)
            length = len(new_block)
            for name in right.keys():
                if name not in left:
                    length += 1
                    if length > max_array_size:
                        raise Error.EvaluateError()
                    new_block[name] = None
            return new_block
        # Result is int or float.
        if Evaluator._is_numbers(left, right):
            result = left % right
            if result % 1 == 0:
                result = int(result)
            return result
        raise Error.EvaluateError()

    @staticmethod
    def _logical_invert(left, right,
                        config=None) -> bool:  # pylint: disable=W0613
        """Invert for "not" operator.

        not true --> false
        not false --> true
        not 0 --> true
        not 10 --> false
        not null --> true
        not "" --> false

        Note that not {"a": 3} is confused as function "not" in Parser.
        So error occurs.
        On the other hand, a = {"a": 3}; return(not a) can be evaluated
        as false.

        Note that not [3] is confused as variable "not" in Parser.
        So error occurs.
        On the other hand, a = [3]; return(not a) can be evaluated
        as false.

        Raises:
            EvaluateError: It is happen if left-hand sided operand is
                           not None.
        """
        if left is None:  # Check just in case even if it is None by Parser.
            return not Evaluator.is_truthy(right)
        raise Error.EvaluateError()

    @staticmethod
    def _logical_and(left, right,
                     config=None) -> bool:  # pylint: disable=W0613
        """Evaluate for "and" operator.

        false and false --> false
        false and true --> false
        true and false --> false
        true and true --> true
        0 and true --> false
        10.5 and true --> true
        "a" and true --> true
        null and true --> false
        true and {"a": 3} --> true
        true and {} --> true
        true and [3] --> true
        true and [] --> true
        """
        return Evaluator.is_truthy(left) and Evaluator.is_truthy(right)

    @staticmethod
    def _logical_or(left, right,
                    config=None) -> bool:  # pylint: disable=W0613
        """Evaluate for "or" operator.

        false or false --> false
        false or true --> true
        true or false --> true
        true or true --> true
        1 or false --> true
        "a" or false --> true
        null or false --> false
        false or {} --> true
        false or {"a": 3} --> true
        false or [] --> true
        false or [3] --> true
        """
        return Evaluator.is_truthy(left) or Evaluator.is_truthy(right)

    @staticmethod
    def _is_equal(left, right, config=None) -> bool:  # pylint: disable=W0613
        """Judge for "==" operator.

        "a" == "a" --> true
        [1, 2] == [1, 2] --> true
        [1, 2] == [2, 1] --> false
        {"a": 1, "b": 2} == {"b": 2, "a": 1} --> true
        {"a": 1, "b": 2} == {"a": 1} --> false
        {"a": 1, "b": 2} == [1, 2] --> false
        {"a": 1, "b": 2} == true --> true (using truthy)
        """
        if Evaluator._is_same_both_type(left, right):
            return left == right
        if isinstance(left, bool) or isinstance(right, bool):
            return Evaluator.is_truthy(left) == Evaluator.is_truthy(right)
        return False

    @staticmethod
    def _is_not_equal(left, right, config=None) -> bool:
        """Judge for "!=" operator."""
        return not Evaluator._is_equal(left, right, config)

    @staticmethod
    def _is_less_than(left, right,
                      config=None) -> bool:  # pylint: disable=W0613
        """Judge for "<" operator.

        Raises:
            EvaluateError: It is happen if the given values are not
                           number.
        """
        Evaluator._make_sure_numbers(left, right)
        return left < right

    @staticmethod
    def _is_less_than_or_equal(left, right,
                               config=None) -> bool:  # pylint: disable=W0613
        """Judge for "<=" operator.

        Raises:
            EvaluateError: It is happen if the given values are not
                           number.
        """
        Evaluator._make_sure_numbers(left, right)
        return left <= right

    @staticmethod
    def _is_greater_than(left, right,
                         config=None) -> bool:  # pylint: disable=W0613
        """Judge for ">" operator.

        Raises:
            EvaluateError: It is happen if the given values are not
                           number.
        """
        Evaluator._make_sure_numbers(left, right)
        return left > right

    @staticmethod
    def _is_greater_than_or_equal(left, right,
                                  config=None) -> bool:  # pylint: disable=W0613
        """Judge for ">=" operator.

        Raises:
            EvaluateError: It is happen if the given values are not
                           number.
        """
        Evaluator._make_sure_numbers(left, right)
        return left >= right

    @staticmethod
    def _is_contained(left, right,
                      config=None) -> bool:  # pylint: disable=W0613
        """Judge for "in" operator.

        "b" in ["a", "b", "c"] --> true
        "d" in ["a", "b", "c"] --> false
        1 in [1, 2, "c"] --> true
        true in [1, true, "c"] --> true
        [2, 3] in [1, [2,3], "c"] --> true
        {"a": 2, "b": false} in [1, {"b": false, "a": 2}, "c"] --> true
        "x" in {"x": 3, "y": 10} --> true
        {"x": 3} in {"x": 3, "y": 10} --> true
        {"x": 3, "y": 10} in {"x": 3, "y": 10} --> true
        "dog" in "cat&dog&bird" --> true
        "23" in "123" --> true
        23 in "123" --> ERROR
        1 in 5 --> ERROR
        [1, 2, "c"] in 1 --> ERROR
        true in true --> ERROR
        null in null --> ERROR

        Raises:
            EvaluateError: It is happen if illegal values are given.
        """
        if isinstance(left, dict) and isinstance(right, dict):
            for name, value in left.items():
                if name not in right or right[name] != value:
                    return False
            return True
        if isinstance(right, str) and not isinstance(left, str):
            raise Error.EvaluateError()
        if isinstance(right, (dict, list, str)):
            return left in right
        raise Error.EvaluateError()

    @staticmethod
    def _remove_element_from_array(elements: list, element):
        """Remove the given element from array."""
        while True:
            try:
                elements.remove(element)
            except ValueError:
                break

    @staticmethod
    def is_truthy(value) -> bool:
        """Evaluate as boolean.

        Args:
            value (Any): A value that is evaluated as boolean.

        Returns:
            bool: The evaluated result
                  - None, False, 0, 0.0 --> False
                  - Other --> True
                    (Note that "0" is True.)
        """
        if value is None:
            return False
        if isinstance(value, bool):
            return value
        if isinstance(value, (int, float)):
            return value != 0
        return True

    @staticmethod
    def _is_same_both_type(left, right) -> bool:
        return type(left) == type(right)  # pylint: disable=C0123

    @staticmethod
    def _is_number(value) -> bool:
        # Note that isinstance(True, int) and isinstance(False, int) is True.
        # So we must check that the given value is not bool.
        return not isinstance(value, bool) and isinstance(value, (int, float))

    @staticmethod
    def _is_numbers(left, right) -> bool:
        return Evaluator._is_number(left) and Evaluator._is_number(right)

    @staticmethod
    def _make_sure_numbers(left, right):
        """Make sure that the arguments are number.

        When the arguments, left and right, aren't number, EvaluateError
        is raised.

        Raises:
            EvaluateError: Exception
        """
        if not Evaluator._is_numbers(left, right):
            raise Error.EvaluateError()

    @staticmethod
    def _make_str_array(values, config=None, is_member=False):
        """Make an array of string with Evaluator.make_str.

        Args:
            values (list[Any]: An array of values.
            config (Config or NoneType, optional): Configuration if
                                                   needed.
            is_member (bool, optional): True if the above value is a
                                        member of block or array,
                                        False otherwise.
                                        The default is False.

        Returns:
            list[str]: an array of string

        Raises:
            EvaluateError: It is happen if number is Infinity or NaN
                           without permission of Configuration.
        """
        strings = []
        for value in values:
            string = Evaluator.make_str(value, config, is_member)
            strings.append(string)
        return strings

    @staticmethod
    def _make_escape_sequence(value: str) -> str:
        """Make escape sequence in string.
        
        Args:
            value (str): The original string

        Returns:
            str: The modified string
        """
        string = value
        index = 0
        length = len(value)
        while index < length:
            letter = string[index]
            if letter in Evaluator._REPLACED_LETTERS:
                string = string[:index] \
                         + '\\' + Evaluator._REPLACED_LETTERS[letter] \
                         + string[index+1:]
                index += 1
                length += 1
            index += 1
        return string

    @staticmethod
    def make_str(value, config=None, is_member=False, is_escape=True
                 ): # pylint: disable=R0911
        """Make string from any value.

        In this class, values are treated as Python native values.
        There are the following difference between VivJson and
        Python.
            - null : None
            - true : True
            - false : False
        Thus, this method is used instead of str function of Python.

        Args:
            value (Any): A value for converting to string.
                         int, float, bool, str, list, dict, None, or
                         Statement.
            config (Config or NoneType, optional): Configuration if
                                                   needed.
            is_member (bool, optional): True if the above value is a
                                        member of block or array,
                                        False otherwise.
                                        The default is False.
            is_escape (bool, optional): True if escape sequence is
                                        enabled, False otherwise.
                                        When this is True, control
                                        character, the double
                                        quotation mark, or reverse
                                        solidus is modified as
                                        escape sequence.
                                        The default is True.

        Returns:
            str: The given value as string

        Raises:
            EvaluateError: It is happen if number is Infinity or NaN
                           without permission of Configuration.
        """
        if value is None:
            return 'null'
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if isinstance(value, (list, tuple)):
            data = Evaluator._make_str_array(value, config, True)
            return f'[{", ".join(data)}]'
        if isinstance(value, dict):
            text = ''
            delimiter = ''
            for name, data in value.items():
                name = Evaluator.make_str(name, config, True)
                data = Evaluator.make_str(data, config, True)
                text = f'{text}{delimiter}{name}: {data}'
                delimiter = ', '
            return f'{{{text}}}'
        if isinstance(value, str):
            if is_escape:
                value = Evaluator._make_escape_sequence(value)
            return f'"{value}"' if is_member else value
        if isinstance(value, float):
            if isinf(value):
                infinity = Config.get_infinity_alt(config)
                if isinstance(infinity, str):
                    sign = '-' if value == float('-inf') else ''
                    return f'{sign}{infinity}'
                raise Error.EvaluateError()
            if isnan(value):
                nan = Config.get_nan_alt(config)
                if nan is not None:
                    return nan
                raise Error.EvaluateError()
        return str(value)



"""Parser for VivJson

- Parser: Constructor. Its argument is source code as string.
- Parser#parse: Parse source code and get statements.

<program> ::= [ "{" [ <nl>+ ] ]
    [ <statement> { <end> <statement> } [ <end> ] ]
    [ [ <nl>+ ] "}" ]
<block> ::= "{" [ <nl>+ ]
    [ <statement> { <end> <statement> } [ <end> ] ]
    [ <nl>+ ] "}"
<statement> ::= "break" | "continue" | <return> | <remove>
    | <expression> | <call_if> | <call_for>
    | <call_primitive> | <call_extended> | <declaration>
<return> ::= "return" [ <nl>+ ] [ "(" [ <nl>+ ] <or> [ <nl>+ ] ")" ]
<remove> ::= "remove" [ <nl>+ ] "(" [ <nl>+ ] <element> [ <nl>+ ] ")"
<declaration> ::= <callee>
<callee> ::= <modifier> [ <nl>+ ] <identifier> [ <nl>+ ]
    [ "(" [ <nl> ] [ <parameter> { <end> <parameter> } ] [ <nl> ] ")" ]
    [ <nl>+ ] <block>
<call_extended> ::= <element> [ <nl>+ ]
    [ "(" [ <nl> ] [ <argument> { <end> <argument> } ] [ <nl> ]
    [ <block> [ <nl> ] ] ")" ] [ <nl>+ ] [ <block> ]
<call_primitive> ::= <element> [ <nl>+ ]
    "(" [ <nl> ] [ <argument> { <end> <argument> } ] [ <nl> ] ")"
<call_for> ::= "for" [ <nl>+ ] "(" [ <nl> ]
    ( [ <argument_for> ] [ <end> ] [ <or> ] [ <end> ]
      [ <argument_for> ] [ <end> ] [ <block> ] )
    [ <nl> ] ")" [ <nl>+ ] [ <block> ]
<call_if> ::= "if" [ <nl>+ ] "(" [ <nl>+ ] <argument>
      [ <end> [ <nl>+ ] <block> ] [ <nl> ] ")" [ <nl>+ ] [ <block> ]
    { [ <nl>+ ] "elseif" [ <nl>+ ] "(" [ <nl>+ ] <argument>
      [ <end> [ <nl>+ ] <block> ] [ <nl> ] ")" [ <nl>+ ] [ <block> ] }
    [ [ <nl>+ ] "else" [ <nl>+ ] [ "("
      [ [ <nl>+ ] <block> ] [ <nl> ] ")" ] [ <nl>+ ] [ <block> ] ]
<parameter> ::= [ <modifier> [ <nl>+ ] ] <identifier>
<argument> ::= <or>
<argument_for> ::= <assignment> | <call_primitive> | <call_extended>
<expression> ::= <assignment> | <result>
<result> ::= ":=" [ <nl>+ ] <or>
<assignment> ::= <element> [ <nl>+ ]
    ( "=" | ":" | "+=" | "-=" | "*=" | "/=" | "%=" ) [ <nl>+ ] <or>
<group> :: = "(" [ <nl>+ ] <or> [ <nl>+ ] ")"
<or> ::= <and> { [ <nl>+ ] "or" [ <nl>+ ] <and> }
<and> ::= <equality> { [ <nl>+ ] "and" [ <nl>+ ] <equality> }
<equality> ::= <comparison>
    { [ <nl>+ ] ( "==" | "!=" | "in" ) [ <nl>+ ] <comparison> }
<comparison> ::= <term>
    { [ <nl>+ ] ( "<" | "<=" | ">" | ">=" ) [ <nl>+ ] <term> }
<term> ::= <factor> { [ <nl>+ ] ( "+" | "-" ) [ <nl>+ ] <factor> }
<factor> ::= <thing>
    { [ <nl>+ ] ( "*" | "/" | "%" ) [ <nl>+ ] <thing> }
<thing> ::= <unary> | <array> | <block>
<unary> ::= [ "+" | "-" | "not" ] ( <primary> | <element>
    | <call_extended> | <call_if> | <call_for> )
<element> ::=
    ( ( <identifier> | <call_primitive> )
      {
        ( [ <nl>+ ] "." [ <nl>+ ]
            ( <identifier> | <call_primitive> |
              ( [ "+" | "-" ] <digit>+ ) ) )
        | ( "[" [ <nl>+ ] <term> [ <nl>+ ] "]" )
      }
    )
<array> :: = "[" [ <nl> ] 
      [ <argument> { <end> <argument> } [ <end> ] ] [ <nl> ]
    "]"
<primary> ::= <number> | <string> | <boolean> | "null" | <group>
<modifier> ::= "function" | "reference"
<identifier> ::= ( "_" | <alphabet> ) { "_" | <alphabet> | <digit> }
<boolean> ::= "true" | "false"
<number> ::= <digit>+ [ "." <digit>+ ]
    [ ( "e" | "E" ) ( "+" | "-" ) <digit>+ ]
<digit> ::= "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
<alphabet> ::= "A" | "B" | "C" | "D" | "E" | "F" | "G" | "H"
    | "I" | "J" | "K" | "L" | "M" | "N" | "O" | "P" | "Q"
    | "R" | "S" | "T" | "U" | "V" | "W" | "X" | "Y" | "Z"
    | "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h"
    | "i" | "j" | "k" | "l" | "m" | "n" | "o" | "p" | "q"
    | "r" | "s" | "t" | "u" | "v" | "w" | "x" | "y" | "z"
<end> ::= <nl> | ";" | ","
<nl> ::= "\n"

<string> is any characters as UTF-8. It is surrounded with "" or ''.

[]: 0 or 1 time (? is not used here. But it has same meaning.)
{}: 0 or more times (* is not used here. But it has same meaning.)
+: 1 or more times
(): group
|: or

"." is allowed as the right-hand sided operand of "in".

Refer to:
- "Let's make a Teeny Tiny compiler"
  https://austinhenley.com/blog/teenytinycompiler2.html
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB sites is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-04-01'

# pylint: disable=C0302, R0903

from .config import Config
from .error import Error
from .lexer import Lexer
from .statement import Literal, Identifier, Keyword, Blank, \
                       Array, Block, Binary, Parameter, \
                       Callee, Call, Get, Set, Remove, Return
from .tokens import Token

class Parser:  # pylint: disable=R0902
    """Parser class

    Attributes:
        _config (Config or NoneType): Configuration
        _is_enable_only_json (bool): When true is given, the given data
                                     is parsed as JSON.
                                     In other words, script is
                                     disabled.
        _lexer (Lexer): An instance of Lexer
        _make_statement_methods (tuple): Methods of making statement.
        _make_unary_methods (tuple): Methods of making unary.
        _make_for_argument_methods (tuple): Methods of making for-loop
                                            argument.
        _tokens (list): Tokens of the current statement.
        _index (int): The current index of the above list "_tokens".
        _is_implicit_assign (bool): When the given source code is not
                                    JSON object (a.k.a. map, hash,
                                    dictionary), this is True.
                                    The default is False.
    """
    def __init__(self, source_code, medium=None, config=None):
        """Initialize class.

        Args:
            source_code (str): Source code as text
            medium (str, optional): The object that has source code.
                                    It is used to report error.
            config (Config, optional): Configuration if needed.
        """
        self._config = config
        self._is_enable_only_json = Config.get_enable_only_json_alt(config)
        self._lexer = Lexer(source_code, medium, config)

        self._make_statement_methods = (
            self._make_return,
            self._make_remove,
            self._make_expression,
            self._make_call_if,
            self._make_call_for,
            self._make_call_extended,
            self._make_call_primitive,
            self._make_declaration
        )
        self._make_unary_methods = (
            self._make_call_for,
            self._make_call_if,
            self._make_call_extended,
            self._make_element,
            self._make_primary
        )
        self._make_for_argument_methods = (
            self._make_block,
            self._make_argument_for,
            self._make_or
        )

        # Initialize tokens.
        self._tokens = []
        self._index = 0
        self._is_implicit_assign = False

    def parse(self):
        """Parse source code and get statements that construct it.

        - Parse VivJson and JSON object (with/without bracket).
        - Parse directly represented value of JSON.

        Returns:
            list[Statement]: <program>

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        # Parse VivJson and JSON object (with/without bracket).
        statements = self._make_program()
        if isinstance(statements, Block):
            return statements.values

        # Parse directly represented value of JSON.
        # It will be an assignment of implicit variable '_'
        return self._parse_directly_represented()

    def _parse_directly_represented(self):
        """Parse directly represented value of JSON.
   
        Target value is number, string, boolean, array, or null.
        The given value is assigned into a implicit variable '_'.
        When some values are given, the value of implicit variable '_'
        is represented as array.
        For example,
          - 3 --> _ = 3
          - [2, 1]; 3 --> _ = [[2, 1], 3]

        Returns:
            list[Statement]: <program>

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        self._is_implicit_assign = True

        arguments = []
        while self._get_token_type() != Token.Type.EOS: # pylint: disable=R1702
            while self._skip_end() is not None:  # Skip <end>+
                pass
            if self._get_token_type() == Token.Type.EOS:
                break

            if self._is_enable_only_json:
                argument = self._make_thing()
                if argument is not None:
                    arguments.append(argument)
                    continue
            else:
                argument = self._make_argument()
                if argument is not None:
                    if isinstance(argument, Get):
                        if len(argument.members) == 1:
                            statement = argument.members[0]
                            if isinstance(statement, Identifier):
                                identifier = statement.name
                                if not self._is_infinity_or_nan(identifier):
                                    string = Token(Token.Type.STRING,
                                                identifier.medium,
                                                identifier.line_number,
                                                identifier.column_number,
                                                identifier.lexeme)
                                    arguments.append(Literal(string))
                                    continue

                    arguments.append(argument)
                    continue

            token_info = self._get_token().to_string(omit_type_name=True)
            self._abort(f'{token_info} is invalid statement/value.')
        if len(arguments) == 0:
            self._abort('No effective statement/value')

        identifier = Identifier(Token(Token.Type.IDENTIFIER, lexeme='_'))
        value = Array(arguments) if len(arguments) > 1 else arguments[0]
        return [Set([identifier], Token(Token.Type.SET), value)]

    def _make_program(self):
        """Make <program>.

        Parse <program> into <statement>.

        <program> ::= [ "{" [ <nl>+ ] ]
            [ <statement> { <end> <statement> } [ <end> ] ]
            [ [ <nl>+ ] "}" ]

        Returns:
            Block or NoneType: <program> as Block if it is obtained,
                               None otherwise.
                               Block is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_block(is_bracket_essential=False, is_consumed=True)

    def _make_block(self, is_bracket_essential=True, is_consumed=False,
                    function_type=Block.ANONYMOUS_FUNCTION):
        """Make <block>.

        <block> ::= "{" [ <nl>+ ]
            [ <statement> { <end> <statement> } [ <end> ] ]
            [ <nl>+ ] "}"

        Args:
            is_bracket_essential (bool, optional): True if left/right
                                    curly brackets are essential.
                                    The default is True.
            is_consumed (bool, optional): When this is True and
                    a statement is constructed, its token are consumed.
                    The default is False.
            function_type (str, optional): Function type of Block
                                  - Block.ANONYMOUS_FUNCTION (default)
                                  - Block.PURE_FUNCTION
                                  - Block.LIMITED_FUNCTION

        Returns:
            Block or NoneType: <block> as Block if it is obtained,
                               None otherwise.
                               Block is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        end_of_block = Token.Type.EOS
        left_bracket = self._match(Token.Type.LEFT_CURLY_BRACKET)  # "{"
        if left_bracket is None:
            if is_bracket_essential:
                return None
        else:
            self._go_to_next_token()
            self._skip_new_lines()  # [ <nl>+ ]
            end_of_block = Token.Type.RIGHT_CURLY_BRACKET  # "}"

        statements = []
        while self._get_token_type() != end_of_block:
            count = len(statements)

            statement = self._make_assignment() if self._is_enable_only_json \
                        else self._make_statement()
            if statement is not None:
                statements.append(statement)

            # <end>s are translated to blank lines.
            # <nl>+ in front of "}" are treated here. Because <end>
            # includes <nl>.
            is_just_behind = statement is not None
            while (token := self._skip_end()) is not None:
                if not is_just_behind:
                    statements.append(Blank(token))
                is_just_behind = False

            # Abort when the effective statement is not found here.
            # However, when the effective statement is completely nothing,
            # we try to assign to implicit variable.
            if count == len(statements):
                if count == 0 and is_consumed \
                        and not self._is_implicit_assign:
                    return None
                token_info = self._get_token().to_string(omit_type_name=True)
                self._abort(f'{token_info} is unexpected.')

            if is_consumed:
                self._consume_token() # Consume Tokens of the above statement.

        if end_of_block == Token.Type.RIGHT_CURLY_BRACKET:
            self._go_to_next_token()

        return Block(statements, function_type)

    def _make_statement(self):
        """Make <statement>.

        <statement> ::= "break" | "continue" | <return> | <remove>
            | <expression> | <call_if> | <call_for>
            | <call_primitive> | <call_extended> | <declaration>

        Returns:
            Statement or NoneType: <statement> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._match(Token.Type.BREAK, Token.Type.CONTINUE)
        if token is not None:
            self._go_to_next_token()
            return Keyword(token)

        for method in self._make_statement_methods:
            statement = method()
            if statement is not None:
                return statement
        return None

    def _make_return(self):
        """Make <return>.

        <return> ::=
            "return" [ <nl>+ ] [ "(" [ <nl>+ ] <or> [ <nl>+ ] ")" ]

        Returns:
            Return or NoneType: <return> as Return if it is
                                obtained, None otherwise.
                                Return is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._match(Token.Type.RETURN)
        if token is None:
            return None
        self._go_to_next_token()

        maybe_restored_index = self._index  # #1

        self._skip_new_lines()  # [ <nl>+ ]

        parenthesis = self._match(Token.Type.LEFT_PARENTHESIS)  # "("
        if parenthesis is None:
            self._index = maybe_restored_index  # Restore #1
            return Return(token, None)
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        value = self._make_or()
        if value is None:
            token_info = parenthesis.to_string(omit_type_name=True)
            self._abort(f'No returned value after {token_info}')

        self._skip_new_lines()  # [ <nl>+ ]

        parenthesis = self._match(Token.Type.RIGHT_PARENTHESIS)  # ")"
        if parenthesis is None:
            token_info = self._get_token().to_string(omit_type_name=True)
            self._abort(f'{token_info} is unexpected.')
        self._go_to_next_token()

        return Return(token, value)

    def _make_remove(self):
        """Make <remove>.

        <remove> ::=
            "remove" [ <nl>+ ] "(" [ <nl>+ ] <element> [ <nl>+ ] ")"

        Returns:
            Remove or NoneType: <remove> as Remove if it is obtained,
                                None otherwise.
                                Remove is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._match(Token.Type.REMOVE)
        if token is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        parenthesis = self._match(Token.Type.LEFT_PARENTHESIS)  # "("
        if parenthesis is None:
            token_info = self._get_token().to_string(omit_type_name=True)
            self._abort(f'{token_info} is unexpected.')
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        element = \
            self._make_element(except_call_primitive=True)  # <element>
        if element is None:
            token_info = parenthesis.to_string(omit_type_name=True)
            self._abort(f'No variable after {token_info}')

        self._skip_new_lines()  # [ <nl>+ ]

        parenthesis = self._match(Token.Type.RIGHT_PARENTHESIS)  # ")"
        if parenthesis is None:
            token_info = self._get_token().to_string(omit_type_name=True)
            self._abort(f'{token_info} is unexpected.')
        self._go_to_next_token()

        return Remove(token, element.members)

    def _make_declaration(self):
        """Make <declaration>.

        <declaration> ::= <callee>

        Returns:
            Callee or NoneType: <declaration> as Callee if it is
                                obtained, None otherwise.
                                Callee is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_callee()

    def _make_callee(self):
        """Make <callee>.

        <callee> ::= <modifier> [ <nl>+ ] <identifier> [ <nl>+ ]
            [ "(" [ <nl> ] [ <parameter> { <end> <parameter> } ]
            [ <nl> ] ")" ] [ <nl>+ ] <block>

        Returns:
            Callee or NoneType: <callee> as Callee if it is
                                obtained, None otherwise.
                                Callee is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1

        # <modifier> [ <nl>+ ] <identifier>
        statement = self._make_parameter(is_modifier_essential=True)
        if statement is None:
            return None

        self._skip_new_lines()  # [ <nl>+ ]

        parameters = self._make_array(Token.Type.LEFT_PARENTHESIS,
                                      Token.Type.RIGHT_PARENTHESIS,
                                      is_parameter=True)
        if parameters is None:
            parameters = Array([])

        self._skip_new_lines()  # [ <nl>+ ]

        block = self._make_block(function_type=Block.PURE_FUNCTION)
        if block is None:
            self._index = maybe_restored_index  # Restore #1
            return None
        parameters.values.append(block)  # It is permitted though frozen=True.

        return Callee(statement, parameters)

    def _make_call_extended(self):
        """Make <call_extended>.

        Returns:
            Call or NoneType: <call_extended> as Call if it is
                              obtained, None otherwise.
                              Call is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_call(is_extended=True)

    def _make_call_primitive(self):
        """Make <call_primitive>.

        Returns:
            Call or NoneType: <call_primitive> as Call if it is
                              obtained, None otherwise.
                              Call is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_call()

    def _make_call(self,  # pylint: disable=R0911
                   is_extended=False, is_keyword=False,
                   function_type=Block.ANONYMOUS_FUNCTION):
        """Make <call_primitive> or <call_extended>.

        <call_primitive> ::= <element> [ <nl>+ ]
            "(" [ <nl> ] [ <argument> { <end> <argument> } ]
                [ <nl> ] ")"

        <call_extended> ::= <element> [ <nl>+ ]
            [ "(" [ <nl> ] [ <argument> { <end> <argument> } ]
                  [ <nl> ] [ <block> [ <nl> ] ] ")" ]
            [ <nl>+ ] [ <block> ]

        Args:
            is_extended (bool, optional): True for <call_extended>,
                                          False for <call_primitive>.
                                          The default is False.
            is_keyword (bool, optional): True if function name's token
                                         type is Keyword,
                                         False otherwise.
                                         The default is False.
            function_type (str, optional): Function type of Block
                                  - Block.ANONYMOUS_FUNCTION (default)
                                  - Block.PURE_FUNCTION
                                  - Block.LIMITED_FUNCTION

        Returns:
            Call or NoneType: <call_primitive> as Call or
                              <call_extended> as Call if it is
                              obtained, None otherwise.
                              Call is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1
        element = None
        if is_keyword:
            element = Identifier(self._get_token())
            self._go_to_next_token()
        else:
            element = \
                self._make_element(except_call_primitive=True)  # <element>
            if element is None:
                return None

        self._skip_new_lines()  # [ <nl>+ ]

        arguments = self._make_array(Token.Type.LEFT_PARENTHESIS,
                                     Token.Type.RIGHT_PARENTHESIS)

        # <call_primitive>
        if not is_extended:
            if arguments is None:
                self._index = maybe_restored_index  # Restore #1
                return None
            return Call(element, arguments)

        # <call_extended>
        self._skip_new_lines()  # [ <nl>+ ]
        block = self._make_block(function_type=function_type)
        if block is None:
            if not isinstance(arguments, Array) \
                    or len(arguments.values) == 0 \
                    or not isinstance(arguments.values[-1], Block):
                self._index = maybe_restored_index  # Restore #1
                return None
            arguments.values[-1] = \
                Block(arguments.values[-1].values, function_type)
            return Call(element, arguments)
        if isinstance(arguments, Array):
            arguments.values.append(block)  # It is permitted though
                                            # frozen=True.
            return Call(element, arguments)
        return Call(element, Array([block]))

    def _make_call_for(self):  # pylint: disable=R0912, R0915
        """Make <call_for>.

        <call_for> ::= "for" [ <nl>+ ] "(" [ <nl> ]
            ( [ <argument_for> ] [ <end> ] [ <or> ] [ <end> ]
              [<argument_for> ] [ <end> ] [ <block> ] )
            [ <nl> ] ")" [ <nl>+ ] [ <block> ]

        Returns:
            Call or NoneType: <call_for> as Call if it is obtained,
                              None otherwise.
                              Call is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        call_for = self._match(Token.Type.FOR)  # "for"
        if call_for is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        bracket = self._match(Token.Type.LEFT_PARENTHESIS)  # "("
        if bracket is None:
            token_info = call_for.to_string(omit_type_name=True)
            self._abort(f'There is no parenthesis after {token_info}.')
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        arguments = []
        flag = 0

        for _ in range(5):
            statement = None
            for method in self._make_for_argument_methods:
                statement = method()
                if statement is not None:
                    arguments.append(statement)
                    break
            end = self._skip_end()  # <end>
            if statement is None:
                if end is not None:
                    arguments.append(Blank(end))
                else:
                    bracket = self._match(Token.Type.RIGHT_PARENTHESIS)  # ")"
                    if bracket is not None:
                        self._go_to_next_token()
                        arguments.append(None)
                        flag |= 1
            elif isinstance(statement, Block):
                flag |= 2
            if flag == 3:
                break

        if len(arguments) < 2:
            token_info = call_for.to_string(omit_type_name=True)
            self._abort(f'There is no operation of {token_info}.' \
                        if len(arguments) < 1 or arguments[0] is None else \
                        f'There is no right parenthesis of {token_info}.')

        block = arguments.pop()
        parenthesis = arguments.pop()
        if block is None and isinstance(parenthesis, Block):
            block = parenthesis
        elif not isinstance(block, Block) or parenthesis is not None:
            token_info = call_for.to_string(omit_type_name=True)
            self._abort(f'Invalid argument in {token_info}')

        length = len(arguments)  # Number of arguments that does not contain
                                 # operations (block).
        if length == 0:  # for () { operation }
            blank = Blank(Token(Token.Type.SEMICOLON))
            arguments = [
                blank,
                Literal(Token(Token.Type.TRUE)),
                blank,
                block
            ]
        elif length == 1:
            if isinstance(arguments[0], Binary) \
                    and arguments[0].operator.type == Token.Type.IN:
                # for (variable in list) { operation }
                variable = arguments[0].left
                if isinstance(variable, Get) and len(variable.members) == 1:
                    variable = variable.members[0]
                if not isinstance(variable, Identifier):
                    token_info = call_for.to_string(omit_type_name=True)
                    self._abort(f'Invalid argument in {token_info}')
                arguments = [
                    Binary(variable, arguments[0].operator,
                           arguments[0].right),
                    block
                ]
            else:  # for (initial) { operation }
                arguments = [
                    arguments[0],
                    Literal(Token(Token.Type.TRUE)),
                    Blank(Token(Token.Type.SEMICOLON)),
                    block
                ]
        elif length == 2:  # for (initial; condition) { operation }
            arguments = [
                arguments[0],
                arguments[1] if not isinstance(arguments[1], Blank) \
                    else Literal(Token(Token.Type.TRUE)),
                Blank(Token(Token.Type.SEMICOLON)),
                block
            ]
        elif length == 3:  # for (initial; condition; update) { operation }
            if isinstance(arguments[1], Blank):
                arguments[1] = Literal(Token(Token.Type.TRUE))
            arguments.append(block)
        else:
            token_info = call_for.to_string(omit_type_name=True)
            self._abort(f'Invalid argument in {token_info}')

        return Call(Identifier(call_for), Array(arguments))

    def _make_call_if(self):
        """Make <call_if>.

        <call_if> ::= "if" [ <nl>+ ] "(" [ <nl>+ ] <argument>
                         [ <end> [ <nl>+ ] <block> ] [ <nl> ] ")"
            [ <nl>+ ][ <block> ]
        { [ <nl>+ ] "elseif" [ <nl>+ ] "(" [ <nl>+ ] <argument>
                         [ <end> [ <nl>+ ] <block> ] [ <nl> ] ")"
            [ <nl>+ ][ <block> ] }
        [ [ <nl>+ ] "else" [ <nl>+ ] [ "("
                         [ [ <nl>+ ] <block> ] [ <nl> ] ")" ]
            [ <nl>+ ] [ <block> ] ]

        Returns:
            Call or NoneType: <call_if> as Call if it is obtained,
                              None otherwise.
                              Call is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        # if
        token = self._match(Token.Type.IF)
        if token is None:
            return None
        condition, block = \
            self._make_call_condition_and_operations(Token.Type.IF)
        if condition is None or block is None:
            return None
        arguments = [condition, block]

        # elseif
        while True:
            condition, block = \
                self._make_call_condition_and_operations(Token.Type.ELSEIF)
            if condition is None or block is None:
                break
            arguments.append(condition)
            arguments.append(block)

        # else
        condition, block = \
            self._make_call_condition_and_operations(Token.Type.ELSE)
        if block is not None:
            arguments.append(Literal(Token(Token.Type.TRUE)))
            arguments.append(block)

        return Call(Identifier(token), Array(arguments))

    def _make_call_condition_and_operations(self, name: Token.Type):
        """Make [(condition)] {operations}.

        - if/elseif (condition) {operations}
        - else {operations}

        Args:
            name (Token.Type): Function name as Token.Type
                               - Token.Type.IF
                               - Token.Type.ELSEIF
                               - Token.Type.ELSE

        Returns:
            Statement or NoneType, Block or NoneType: Condition and
                            Operations if they are obtained.
                            None and None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._match(name)
        if token is None:
            return None, None

        length = 1 if name == Token.Type.ELSE else 2

        call = self._make_call(is_extended=True, is_keyword=True,
                               function_type= Block.LIMITED_FUNCTION)
        # pylint: disable=E1101
        if not isinstance(call, Call) \
                or not isinstance(call.name, Identifier) \
                or call.name.name.type != name \
                or len(call.arguments.values) < length \
                or not isinstance(call.arguments.values[-1], Block):
            token_info = token.to_string(omit_type_name=True)
            self._abort(f'{token_info} is invalid.')

        arguments = []
        for statement in call.arguments.values:
            if isinstance(statement, Blank):
                continue

            if len(arguments) < length:
                arguments.append(statement)
            else:
                arguments = []  # It indicates error.
                break

        if len(arguments) != length or not isinstance(arguments[-1], Block):
            token_info = token.to_string(omit_type_name=True)
            self._abort(f'{token_info} is invalid.')
        if length == 1:
            return None, arguments[0]
        return arguments[0], arguments[1]

    def _make_parameter(self, is_modifier_essential=False):
        """Make <parameter>.

        <parameter> ::= [ <modifier> [ <nl>+ ] ] <identifier>
        <modifier> ::= "function" | "reference"

        Args:
            is_modifier_essential (bool, optional): True if <modifier>
                                                    is essential.

        Returns:
            Parameter or NoneType: <parameter> as Parameter if it is
                                    obtained, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1
        modifier = self._match( # Token.Type.CLASS,
                               Token.Type.FUNCTION,
                               Token.Type.REFERENCE)
        if modifier is None:
            if is_modifier_essential:
                return None
        else:
            self._go_to_next_token()
            self._skip_new_lines()  # [ <nl>+ ]

        identifier = self._match(Token.Type.IDENTIFIER)
        if identifier is None:
            if modifier is not None:
                token_info = modifier.to_string(omit_type_name=True)
                self._abort(f'Unexpected statement after {token_info}')
            self._index = maybe_restored_index  # Restore #1
            return None
        self._go_to_next_token()

        return Parameter(modifier, Identifier(identifier))

    def _make_argument(self):
        """Make <argument>.

        <argument> ::= <or>

        Returns:
            Statement or NoneType: <or> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_or()

    def _make_argument_for(self):
        """Make <argument_for>.

        <argument_for> ::= <assignment> | <call_primitive>
            | <call_extended>

        Returns:
            Statement or NoneType: <argument_for> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        statement = self._make_assignment()
        if statement is None:
            statement = self._make_call_primitive()
            if statement is None:
                statement = self._make_call_extended()
        return statement

    def _make_expression(self):
        """Make <expression>.

        <expression> ::= <assignment> | <result>

        Returns:
            Statement or NoneType: <expression> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        statement = self._make_assignment()
        if statement is None:
            statement = self._make_result()
        return statement

    def _make_assignment(self):
        """Make <assignment>.

        <assignment> ::= <element> [ <nl>+ ]
            ( "=" | ":" | "+=" | "-=" | "*=" | "/=" | "%=" ) [ <nl>+ ]
            <or>

        Returns:
            Set or NoneType: <assignment> if it is obtained,
                             None otherwise.
                             Set is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1

        element = None
        if not self._is_enable_only_json:
            element = self._make_element()
            if element is not None:
                # _make_element method returns an instance of Get data
                # class.
                # But this method makes an instance of Set data class.
                # Therefore, it is extracted.
                element = element.members

        if element is None:
            # Allow string that is surrounded with quotation marks,
            # such as {'key': 100}.
            token = self._match(Token.Type.IDENTIFIER, Token.Type.STRING) \
                    if self._is_enable_only_json \
                    else self._match(Token.Type.STRING)
            if token is None:
                return None
            token.type = Token.Type.IDENTIFIER
            element = [Identifier(token)]
            self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        operator = self._match(Token.Type.SET) if self._is_enable_only_json \
                   else self._match(
                    Token.Type.ASSIGN, Token.Type.SET,
                    Token.Type.PLUS_ASSIGN, Token.Type.MINUS_ASSIGN,
                    Token.Type.MULTIPLY_ASSIGN, Token.Type.DIVIDE_ASSIGN,
                    Token.Type.MODULO_ASSIGN)
        if operator is None:
            self._index = maybe_restored_index  # Restore #1
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        argument = self._make_thing() if self._is_enable_only_json \
                   else self._make_or()  # <thing> or <or>
        if argument is None:
            self._index = maybe_restored_index  # Restore #1
            return None

        return Set(element, operator, argument)

    def _make_result(self):
        """Make <result>.

        <result> ::= ":=" [ <nl>+ ] <or>

        Returns:
            Set or NoneType: <result> as Set if it is obtained,
                             None otherwise.
                             Set is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1
        operator = self._match(Token.Type.RESULT)  # ":="
        if operator is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        argument = self._make_or()  # <or>
        if argument is None:
            self._index = maybe_restored_index  # Restore #1
            return None

        return Set([], operator, argument)

    def _make_group(self):
        """Make <group>.

        <group> :: = "(" [ <nl>+ ] <or> [ <nl>+ ] ")"

        Returns:
            Statement or NoneType: <or> or <group> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1
        token = self._match(Token.Type.LEFT_PARENTHESIS)  # "("
        if token is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        statement = self._make_or()  # <or>
        if statement is None:
            token_info = token.to_string(omit_type_name=True)
            self._abort(f'Bad statement after {token_info}')

        self._skip_new_lines()  # [ <nl>+ ]

        token = self._match(Token.Type.RIGHT_PARENTHESIS)  # ")"
        if token is None:
            self._index = maybe_restored_index  # Restore #1
            return None
        self._go_to_next_token()

        return statement

    def _make_or(self):
        """Make <or>.

        <or> ::= <and> { [ <nl>+ ] "or" [ <nl>+ ] <and> }

        Returns:
            Statement or NoneType: <and> or <or> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary(Token.Type.OR, self._make_and)

    def _make_and(self):
        """Make <and>.

        <and> ::= <equality> { [ <nl>+ ] "and" [ <nl>+ ] <equality> }

        Returns:
            Statement or NoneType: <equality> or <and> if it is
                                   obtained, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary(Token.Type.AND, self._make_equality)

    def _make_equality(self):
        """Make <equality>.

        <equality> ::= <comparison>
            { [ <nl>+ ] ( "==" | "!=" | "in" ) [ <nl>+ ] <comparison> }

        Returns:
            Statement or NoneType: <comparison> or <equality> if it is
                                   obtained, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary(
                    (Token.Type.EQ, Token.Type.NE, Token.Type.IN),
                    self._make_comparison)

    def _make_comparison(self):
        """Make <comparison>.

        <comparison> ::= <term>
            { [ <nl>+ ] ( "<" | "<=" | ">" | ">=" ) [ <nl>+ ] <term> }

        Returns:
            Statement or NoneType: <term> or <comparison> as Binary
                                   if it is obtained, None
                                   otherwise.
                                   Binary is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary(
            (Token.Type.LT, Token.Type.LE, Token.Type.GT, Token.Type.GE),
            self._make_term)

    def _make_term(self):
        """Make <term>.

        <term> ::=
            <factor> { [ <nl>+ ] ( "+" | "-" ) [ <nl>+ ] <factor> }

        Returns:
            Statement or NoneType: <factor> or <term> as Binary if it
                                   is obtained, None otherwise.
                                   Binary is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary((Token.Type.PLUS, Token.Type.MINUS),
                                 self._make_factor)

    def _make_factor(self):
        """Make <factor>.

        <factor> ::=
            <thing> { [ <nl>+ ] ( "*" | "/" | "%" ) [ <nl>+ ] <thing> }

        Returns:
            Statement or NoneType: <thing> or <factor> as Binary if it
                                   is obtained, None otherwise.
                                   Binary is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        return self._make_binary(
            (Token.Type.MULTIPLY, Token.Type.DIVIDE, Token.Type.MODULO),
            self._make_thing)

    def _make_thing(self):
        """Make <thing>.

        <thing> ::= <unary> | <array> | <block>

        Returns:
            Statement or NoneType: <unary> or <array> or <block> if it
                                   is obtained, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        statement = self._make_unary()
        if statement is None:
            statement = self._make_array()
            if statement is None:
                statement = self._make_block()
        return statement

    def _make_binary(self, operator_tokens, *method_for_making_statement):
        """Make binary.

        binary ::= <statement> 
            { [ <nl>+ ] <operator> [ <nl>+ ] <statement> }

        Args:
            operator_tokens (tuple): Tokens of operator
            method_for_making_statement (function): Method for making
                    sub class of Statement.
                    When 2 methods are given, 1st method is used for
                    left side and other is used for right side.
                    When only 1 method is given, it is used for both
                    sides.

        Returns:
            Statement or NoneType: <statement> or Binary if it is
                                   obtained, None otherwise.
                                   Binary is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        statement = method_for_making_statement[0]()  # <statement>
        if statement is None:
            return None

        while True:  # {}
            maybe_restored_index = self._index  # #1

            self._skip_new_lines()  # [ <nl>+ ]

            operator = self._match(operator_tokens)  # <operator>
            if operator is None:
                break
            self._go_to_next_token()

            self._skip_new_lines()  # [ <nl>+ ]

            right = method_for_making_statement[-1]()  # <statement>
            if right is None:
                if operator.type == Token.Type.IN:
                    dot = self._match(Token.Type.DOT)
                    if dot is not None:
                        dot.type = Token.Type.IDENTIFIER
                        right = Identifier(dot)
                        self._go_to_next_token()
                if right is None:
                    token_info = operator.to_string(omit_type_name=True)
                    self._abort(f'Bad statement after {token_info}')

            statement = Binary(statement, operator, right)

        self._index = maybe_restored_index  # Restore #1
        return statement

    def _make_unary(self):
        """Make <unary>.

        <unary> ::= [ "+" | "-" | "not" ] ( <primary> | <element>
            | <call_extended> | <call_if> | <call_for> )

        Returns:
            Statement or NoneType: <primary>, <element>,
                                   <call_extended>, <call_if>,
                                   <call_for>, or <unary> if it is
                                   obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        maybe_restored_index = self._index  # #1
        prefix = self._match(Token.Type.PLUS, Token.Type.MINUS) \
                 if self._is_enable_only_json \
                 else self._match(Token.Type.PLUS, Token.Type.MINUS,
                                  Token.Type.NOT)
        if prefix is not None:
            self._go_to_next_token()

        statement = None
        if self._is_enable_only_json:
            statement = self._make_primary()
        else:
            for method in self._make_unary_methods:
                statement = method()
                if statement is not None:
                    break
        if statement is None:
            self._index = maybe_restored_index  # Restore #1
            return None

        if prefix is None or prefix.type == Token.Type.PLUS:
            return statement
        if prefix.type == Token.Type.MINUS:
            return Binary(Literal(Token(Token.Type.NUMBER,
                                        medium=prefix.medium,
                                        line_number=prefix.line_number,
                                        column_number=prefix.column_number,
                                        lexeme='-1')),
                          Token(Token.Type.MULTIPLY), statement)
        return Binary(Literal(Token(Token.Type.NULL)), prefix, statement)

    def _make_element(self, except_call_primitive=False):
        """Make <element>.

        <element> ::=
        ( ( <identifier> | <call_primitive> )
          {
            ( [ <nl>+ ] "." [ <nl>+ ]
                ( <identifier> | <call_primitive> | <digit>+ ) )
            | ( "[" [ <nl>+ ] <unary> [ <nl>+ ] "]" )
          }
        )

        Args:
            except_call_primitive (bool, optional): True if
                                    <call_primitive> is not allowed.
                                    The default is False.

        Returns:
            Get or NoneType: <element> as Get if it is obtained,
                             None otherwise.
                             Get is sub class of Statement

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        # <identifier> | <call_primitive>
        member = None if except_call_primitive \
                    else self._make_call_primitive()  # <call_primitive>
        if member is None:
            token = self._match(Token.Type.IDENTIFIER)
            if token is None:
                return None
            member = Identifier(token)  # <identifier>
            self._go_to_next_token()

        members = [member]

        while True:  # {}
            maybe_restored_index = self._index  # #1

            # "[" [ <nl>+ ] <term> [ <nl>+ ] "]"
            member = self._make_member_with_bracket()
            if member is not None:
                members.append(member)
                continue

            # [ <nl>+ ] "." [ <nl>+ ]
            # ( <identifier> | <call_primitive> | ( [ "+" | "-" ] <digit>+ ) )
            member = self._make_member_with_dot()
            if member is not None:
                members.append(member[0])
                if member[1] is not None:
                    members.append(member[1])
                continue

            self._index = maybe_restored_index  # Restore #1
            break

        return Get(members)

    def _make_member_with_bracket(self):
        """Make bracket member, such as *[a], *[3].

        "[" [ <nl>+ ] <term> [ <nl>+ ] "]"

        Returns:
            Statement or NoneType: <unary> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        bracket = self._match(Token.Type.LEFT_SQUARE_BRACKET)  # "["
        if bracket is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        member = self._make_term()  # <term>
        if member is None:
            token_info = bracket.to_string(omit_type_name=True)
            self._abort(f'Unexpected statement after {token_info}.')

        self._skip_new_lines()  # [ <nl>+ ]

        bracket = self._match(Token.Type.RIGHT_SQUARE_BRACKET)  # "]"
        if bracket is None:
            token_info = self._get_token().to_string(omit_type_name=True)
            self._abort(f'{token_info} is unexpected.')
        self._go_to_next_token()
        return member

    def _make_member_with_dot(self):
        """Make dot member, such as *.a, *.3, *.func().

        [ <nl>+ ] "." [ <nl>+ ]
        ( <identifier> | <call_primitive> |
            ( [ "+" | "-" ] <digit>+ )
        )

        The returned value is a list that has 2 statements.
        However 2nd statement is null mostly. 2nd statement becomes
        valid statement if it seems that they are floating-point
        number.
        There are some example in the following table. To tell the
        truth, in the last item "f.0.2", 0.2 is entered in a Token
        that is created by Lexer. Thus, this parsing is needed.

        | Target value                 | Expression | 1st | 2nd  |
        |------------------------------|------------|-----|------|
        | {"a": {"x": 3}}              | a.x        | "x" | null |
        | {"b": [7, 8, 9]}             | b.0        |  0  | null |
        | {"c": [10, [50, 70]]}        | c.1.0      |  1  |  0   |
        | {"d": {"2": {"6": 100}}}     | d.2.6      |  2  |  6   |
        | {"e": {"3": [1000, 2000]}}   | e.3.1      |  3  |  1   |
        | {"f": [{"2": 10000}, 20000]} | f.0.2      |  0  |  2   |

        Returns:
            list[Statement] or NoneType: <identifier>,
                                         <call_primitive>, or
                                         <digit>+ if it is obtained,
                                         None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        self._skip_new_lines()  # [ <nl>+ ]

        dot = self._match(Token.Type.DOT)
        if dot is None:
            return None
        self._go_to_next_token()

        self._skip_new_lines()  # [ <nl>+ ]

        member = self._make_call_primitive()  # <call_primitive>
        additional_member = None
        if member is None:
            token = self._match(Token.Type.IDENTIFIER, Token.Type.STRING)
            if token is not None:
                # Literal data class is used instead of Identifier data class.
                # It prevents this member from evaluation as variable.
                token.type = Token.Type.STRING
                member = Literal(token)  # <identifier>
            else:
                prefix = self._match(Token.Type.PLUS, Token.Type.MINUS)
                if prefix is not None:
                    self._go_to_next_token()
                token = self._match(Token.Type.NUMBER)
                if token is None:
                    token_info = dot.to_string(omit_type_name=True)
                    self._abort(f'Unexpected member after {token_info}')
                if '.' not in token.lexeme:
                    member = Literal(token)  # <digit>+
                else:
                    numbers = token.lexeme.split('.', maxsplit=2)
                    # The range of len(numbers) may be 0-3. But it must be 2.
                    if len(numbers) != 2 or not numbers[1].isdigit():
                        token_info = token.to_string(omit_type_name=True)
                        self._abort(f'Unexpected member after {token_info}')
                    for index, number in enumerate(numbers):
                        numbers[index] = Literal(
                                Token(Token.Type.NUMBER, medium=token.medium,
                                      line_number=token.line_number,
                                      column_number=token.column_number,
                                      lexeme=number))
                    member = numbers[0]
                    additional_member = numbers[1]
                if prefix is not None and prefix.type == Token.Type.MINUS:
                    member = Binary(
                              Literal(Token(Token.Type.NUMBER,
                                           medium=prefix.medium,
                                           line_number=prefix.line_number,
                                           column_number=prefix.column_number,
                                           lexeme='-1')),
                              Token(Token.Type.MULTIPLY), member)
            self._go_to_next_token()
        return [member, additional_member]

    def _make_array(self,
                    left_bracket=Token.Type.LEFT_SQUARE_BRACKET,
                    right_bracket=Token.Type.RIGHT_SQUARE_BRACKET,
                    is_parameter=False):
        """Make <array>.

        Normally, <array> :: = "[" [ <nl> ] [ <argument>
        { <end> <argument> } [ <end> ] ] [ <nl> ] "]" is made.

        "(" and ")" can be used instead of "[" and "]" for caller.
        <parameter> can be used instead of <argument> for callee.

        Args:
            left_bracket (Toke.Type, optional): The left side of
                                                surrounding array.
                                                The default is "[".
            right_bracket (Toke.Type, optional): The right side of
                                                 surrounding array.
                                                 The default is "]".
            is_parameter (bool, optional): True for using <parameter>
                                           instead of <argument>,
                                           False otherwise.
                                           The default is False.

        Returns:
            Array or NoneType: <array> as Array if it is obtained,
                               None otherwise.
                               Array is sub class of Statement.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        bracket = self._match(left_bracket)
        if bracket is None:
            return None
        self._go_to_next_token()

        self._skip_new_line()  # [ <nl> ]

        values = []
        while self._get_token_type() != right_bracket:
            value = self._make_argument() if not is_parameter \
                        else self._make_parameter()
            if value is not None:
                values.append(value)

            end = self._skip_end()
            if value is None:
                if end is None:
                    # There is no right bracket, no value, no ";",
                    # no ",", and no new line code. It is invalid.
                    token_info = \
                        self._get_token().to_string(omit_type_name=True)
                    self._abort(f'{token_info} is unexpected.')

                if end.type == Token.Type.NEW_LINE:
                    # Blank line is ignored.
                    continue

                # There is no value though delimiter is existed.
                # However it is permitted for "for (;;)",
                # for (; i < 10; i += 1), and so on.
                # In other words, it is not permitted for parameter of
                # function definition and array.
                if is_parameter:  # Parameter of function definition
                    token_info = end.to_string(omit_type_name=True)
                    self._abort(
                        f'{token_info} is found without parameter.')
                if left_bracket == Token.Type.LEFT_SQUARE_BRACKET:  # array
                    token_info = end.to_string(omit_type_name=True)
                    self._abort(f'{token_info} is found without value.')

        self._go_to_next_token()

        return Array(values)

    def _make_primary(self):
        """Make <primary>.

        <primary> ::= <number> | <string> | <boolean> | "null"
            | <group>

        Returns:
            Statement or NoneType: <primary> if it is obtained,
                                   None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._match(Token.Type.NUMBER, Token.Type.STRING,
                            Token.Type.TRUE, Token.Type.FALSE,
                            Token.Type.NULL)

        if token is None:
            token = self._match(Token.Type.IDENTIFIER)
            if token is not None:
                if self._is_infinity_or_nan(token):
                    self._go_to_next_token()
                    return Identifier(token)
                if self._is_implicit_assign:
                    token.type = Token.Type.STRING
                else:
                    token = None

        if token is not None:
            self._go_to_next_token()
            return Literal(token)

        if self._is_enable_only_json:
            return None

        return self._make_group()

    def _is_infinity_or_nan(self, token: Token) -> bool:
        """Detect whether the given token is Infinity/NaN or not.

        Args:
            token (Token): Token

        Returns:
            bool: True if the given token is Infinity/NaN,
                  False otherwise
        """
        identifiers = [Config.get_infinity_alt(self._config),
                       Config.get_nan_alt(self._config)]
        for identifier in identifiers:
            if identifier is not None and identifier == token.lexeme:
                return True
        return False

    def _match(self, *token_types):
        """Get token if the given types have the current token type.

        Example:
        - self._match(Token.Type.NUMBER)
        - self._match(Token.Type.NUMBER, Token.Type.STRING)
        - self._match((Token.Type.NUMBER, Token.Type.STRING))
        - self._match([Token.Type.NUMBER, Token.Type.STRING])

        Args:
            token_types (tuple or list or Token.Type): Token types.

        Returns:
            Token or NoneType: Token if the current token type is
                    existed in the given types, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: It is happen when parsing is failed.
        """
        token = self._get_token()
        if len(token_types) > 1:  # multiple parameters
            if token.type not in token_types:
                token = None
        elif isinstance(token_types[0], (list, tuple)):
            if token.type not in token_types[0]:
                token = None
        elif isinstance(token_types[0], Token.Type):
            if token.type != token_types[0]:
                token = None
        else:
            self._abort("Interpreter side error in Parser#_match")
        return token

    def _go_to_next_token(self):
        """Process a next token.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        self._index += 1
        self._get_token()

    def _get_token_type(self):
        """Get current Token type.

        Returns:
            Token.Type: The current Token type

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        return self._get_token().type

    def _get_token(self):
        """Get current Token.

        Returns:
            Token: The current Token

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        if self._index > len(self._tokens):
            self._abort("Interpreter side error in Parser#_get_token")

        if self._index == len(self._tokens):
            self._tokens.append(self._lexer.get_token())
        return self._tokens[self._index]

    def _consume_token(self):
        """Consume the current token.

        For example, the current token is "abc" and it is consumed.
        self._index points the head of next token.

                                    self._index
                                    |
                        0   1   2   3   4   5   6
                      +---+---+---+---+---+---+---+
        self._tokens: | a | b | c | d | e | f | g |
                      +---+---+---+---+---+---+---+

                             |
                             V

                        self._index
                        |
                        0   1   2   3
                      +---+---+---+---+
        self._tokens: | d | e | f | g |
                      +---+---+---+---+
        """
        if self._index > len(self._tokens):
            self._abort("Interpreter side error in Parser#_consume_token")

        # Remove from 0 to (self._index - 1).
        del self._tokens[0:self._index]
        self._index = 0

    def _skip_new_line(self):
        """Skip <nl>.

        Skip new line code from the current position.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        if self._get_token_type() == Token.Type.NEW_LINE:
            self._go_to_next_token()

    def _skip_new_lines(self):
        """Skip <nl>+.

        Skip new line codes from the current position.
        As a result, there is other letter in the current position.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        while self._get_token_type() == Token.Type.NEW_LINE:
            self._go_to_next_token()

    def _skip_end(self):
        """Skip <end>.

        Skip the following tokens from the current position.
        - ";" "\n"
        - "," "\n"
        - "\n"

        Returns:
            Token or NoneType: The skipped <end> Token if <end> is
                               skipped, None otherwise.

        Raises:
            LexError: It is happen when unknown token is found.
            ParseError: The out of range is happen.
        """
        is_end = False

        token = self._get_token()
        if token.type in (Token.Type.SEMICOLON, Token.Type.COMMA):
            is_end = True
            self._go_to_next_token()

        if self._get_token_type() == Token.Type.NEW_LINE:
            is_end = True
            self._go_to_next_token()

        return token if is_end else None

    def _abort(self, message):
        """Abort Parser.

        Print an error message and raise Exception.

        Args:
            message (str): Error message

        Raises:
            ParseError: This exception always happen when this method
                        is called.
        """
        raise Error.make_parse_error(message, config=self._config)



"""Unit-test for Standard library

Run `python3 -m pytest tests/test_standard.py` in parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-30'

# pylint: disable=C0301, E0401, R0801

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        ('a = 10; if (a > 5) { a += 100 }', 'a', 110, False),
        ('a = 10; if (a > 5) { a += 100 } else { a -= 100 }', 'a', 110, False),
        ('a = 1; if (a > 5) { a += 100 } else { a -= 100 }', 'a', -99, False),
        ('a = 1; if (a > 5) { a += 100 } elseif (a < 0) { a -= 100 }', 'a', 1, False),
        ('x = if (true) { a = 100 } else { a = -100 }', 'x', {"a": 100}, False),
        ('x = if (true) { := 100 } else { := -100 }', 'x', 100, False),
        ('a = -1, x = if (a < 0) {:= "-"} elseif (a == 0) {:="zero"} else {:="+"}', 'x', "-", False),
        ('a = 0, x = if (a < 0) {:= "-"} elseif (a == 0) {:="0" :="zero" x=0} else {:="+"}', 'x', "zero", False),
        ('a = 1.5, x = if (a < 0) {:= "-"} elseif (a == 0) {:="zero"} else {:="+"}', 'x', "+", False),
        ('a = -0.01, b = -1, result = "" ' \
         'if (a < 0) { if (b < 0) { result = "< <" } else { result = "< >=" } } ' \
         'else { if (b < 0) { result = ">= <" } else { result = ">= >=" } }',
         'result', "< <", False),
        ('a = -10, b = 100000, result = "" ' \
         'if (a < 0) { if (b < 0) { result = "< <" } else { result = "< >=" } } ' \
         'else { if (b < 0) { result = ">= <" } else { result = ">= >=" } }',
         'result', "< >=", False),
        ('a = 0, b = -1e-3, result = "" ' \
         'if (a < 0) { if (b < 0) { result = "< <" } else { result = "< >=" } } ' \
         'else { if (b < 0) { result = ">= <" } else { result = ">= >=" } }',
         'result', ">= <", False),
        ('a = -1, b = -1e-8, result = "" ' \
         'if (a < 0) { if (b < 0) { result = "< <" } else { result = "< >=" } } ' \
         'else { if (b < 0) { result = ">= <" } else { result = ">= >=" } }',
         'result', "< <", False),
        ('a = -10, b = 100000, result = "" ' \
         'if (a < 0) { if (b < 0) { result = "< <" } /* else { result = "< >=" } */ } ' \
         'else { if (b < 0) { result = ">= <" } else { result = ">= >=" } }',
         'result', "", False),
        ('a = -0.01, b = -1, result = "", ' \
         'if (a < 0 and b < 0) { result = "< <" } '\
         'elseif (a < 0 and b >= 0) { result = "< >=" } ' \
         'elseif (a >= 0 and b < 0) { result = ">= <" } ' \
         'else { result = ">= >=" }',
         'result', "< <", False),

        ('a = [] do {a += 1, a += 2}', 'a', [1, 2], False),
        ('a = [] do {a += 1, break, a += 2}', 'a', [1], False),
        ('i=0, do {i += 1, continue}', 'i', None, True),
        ('i=0, do {i += 1, if (i <= 10) {continue}}', 'i', 11, False),
        ('x = do {i = 100, a = true, b = null}', 'x', {"i": 100, "a": True, "b": None}, False),

        ('i = 0; while (i < 10) { a = i;  i += 1}', 'i', 10, False),
        ('_i = 0; while (_i < 10) { a = _i;  _i += 1}', '_i', 10, False),
        ('i = 0; while (true) { i+=1 }', 'i', None, True),  # Infinite loop
        ('i = 0; while (true) { i+=1 if (i >= 3) {break} }', 'i', 3, False),
        ('i = 0, x = [], while (i < 10) { i += 1 if (i < 5) {continue} x += i }', 'x', [5, 6, 7, 8, 9, 10], False),
        ('i = 1 a = while (i < 10) { return(100) }', '[i, a]', 100, False),  # 100 is returned because {i = 1, a = while (i < 10) { return(100) }, return([i, a])}
        ('i = 1 a = while (i < 10) { i += 1 }', '[i, a]', [10, {}], False),
        ('i = 1 a = while (i < 10 { i += 1 })', '[i, a]', [10, {}], False),
        ('i = 1; a = while (i < 10; { i += 1 })', '[i, a]', [10, {}], False),
        ('i = 1 a = while (i < 10) { last = i, i += 1 }', '[i, a]', [10, {"last": 9}], False),
        ('i = 1 a = while (i < 10) { := i, i += 1 }', '[i, a]', [10, 9], False),

        ('i = 10; for (; i >= 0; i -= 2) { a = i }', 'i', -2, False),
        ('for (i = 0; i < 20; i += 1) { a = i; }', '[a, i]', [None, None], False),
        ('a = null, for (i = 0; i < 20; ; ; ) { i += 1, a = i }', '[i, a]', None, True),  # Arguments is too much.
        ('a = null, for (i = 0; i < 20; ; ) { i += 1, a = i }', '[i, a]', [None, 20], False),
        ('a = null, for (i = 0; i < 20; ) { i += 1, a = i }', '[i, a]', [None, 20], False),
        ('a = null, for (i = 0; i < 20) { i += 1, a = i }', '[i, a]', [None, 20], False),
        ('i = 0; for () { i+=1 }', 'i', None, True),  # Infinite loop
        ('i = 0; for (;;) { i+=1 }', 'i', None, True),  # Infinite loop
        ('i = 0; for (;;i+=1) { i+=0.5 }', 'i', None, True),  # Infinite loop
        ('i = 0; for (;;) { i = 1 break i = 2 }', 'i', 1, False),
        ('i = 0; for (;;) { i += 1 continue i = 2 break }', 'i', None, True),  # Infinite loop
        ('i = 1 a = for (; i < 10; i += 1) { return(100) }', '[i, a]', 100, False),  # 100 is returned because {i = 1, a = for (; i < 10; i += 1) { return(100) }, return([i, a])}
        ('i = 1 a = for (; i < 10; i += 1; { return(100) })', '[i, a]', 100, False),  # 100 is returned because {i = 1, a = for (; i < 10; i += 1; { return(100) }), return([i, a])}
        ('a = for (i = 0; i < 10; i += 1) {}', '[i, a]', [None, {"i": 10}], False),
        ('a = for (i = 0; i < 10; i += 1) {:=false}', '[i, a]', [None, False], False),
        ('i = null, for (i = 0) {break}', 'i', 0, False),
        ('i = 0, for (i < 0) {break}', 'i', 0, False),

        ('x = [], for (a in [3, 5, 1]) {x += a * 2}', 'x', [6, 10, 2], False),
        ('x = "", for (a in ["foo", "bar", "baz"]) {x += a + "," }', 'x', "foo,bar,baz,", False),
        ('x = [], list = [3, 5, 1], for (a in list) {x += a * 2}', '[x, list]', [[6, 10, 2], [3, 5, 1]], False),
        ('x = {}, list = {x: 3, y: 5, z: 1}, for (a in list) {x[a[0]] = a[1]*2}', 'x', {"x": 6, "y": 10, "z": 2}, False),
        ('list = [{x: 3, y: 5}], z = null, for (a in list) {z = a.x}', 'z', 3, False),
        ('list = [{x: 3, y: 5}], for (a in list) {a.x = 2}', 'list.0.x', 3, False),
        ('list = [{x: 3, y: 5}], z = null, for (a in list) {a.x = 2, z = a.x}', 'z', 2, False),
        ('list = [{x: 3, y: 5}], for (a in list) {a.x = 2}', 'list', [{"x": 3, "y": 5}], False),
        ('list = {"x": {"y": 5}}, for (a in list) {a[1]["y"] = 1}', 'list', {"x": {"y": 5}}, False),
        ('x = [], list = 1, for (a in list) {x += a * 2}', 'x', None, True),  # Array or Block is permitted.
        ('"foo": 10, "bar": 30, "baz": 20, max=-1, for (pair in .) {if (max < pair[1]) {max = pair[1]}}', 'max', 30, False),

        ('', '[1.0, int(1.0), int(2.8), int("10.3"), int("1e3")]', [1.0, 1, 2, 10, 1000], False),
        ('x = "3.5"', 'int(x)', 3, False),
        ('', 'int()', None, True),  # The argument is necessary.
        ('', 'int("x")', None, True),  # The argument must be number.
        ('', 'int(true)', None, True),  # The argument must be number.
        ('', 'int([1])', None, True),  # The argument must be number.
        ({"a": float('nan')}, 'int(a)', None, True),  # NaN can not be converted.
        ({"a": float('inf')}, 'int(a)', None, True),  # Infinity can not be converted.

        ('', '[10, float(10)]', [10, 10.0], False),
        ('x = "3.5"', 'float(x)', 3.5, False),
        ('', 'float()', None, True),  # The argument is necessary.
        ('', 'float("x")', None, True),  # The argument must be number.
        ('', 'float(true)', None, True),  # The argument must be number.

        ('', 'string(3 + 2)', '5', False),
        ('x = 1.5', 'string(x)', '1.5', False),
        ('', 'string(true)', 'true', False),
        ('', 'string(false)', 'false', False),
        ('', 'string(null)', 'null', False),
        ('', 'string()', None, True),  # The argument is necessary.

        ('', 'len("")', 0, False),
        ('', 'len("abc")', 3, False),
        ('a = len("xyz")', 'a', 3, False),
        ('a = "xyz"', 'len(a)', 3, False),
        ('a = "αβz"', 'len(a)', 3, False),  # multi-byte character "αβ"
        ('', 'len([])', 0, False),
        ('', 'len(["abc", "12345"])', 2, False),
        ('', 'len({})', 0, False),
        ('', 'len({"a": 1, "b": true, "c": null})', 3, False),
        ('', 'len()', None, True),  # The argument is necessary.
        ('', 'len(2)', None, True),  # The argument must be array, block, string
        ('', 'len(true)', None, True),  # The argument must be array, block, string
        ('', 'len(null)', None, True),  # The argument must be array, block, string

        ('a = [], insert(a, 0, 10)', 'a', [10], False),
        ('a = [1, 2, 3, 4, 5], insert(a, len(a), [null, true, false, "text"])', 'a', [1, 2, 3, 4, 5, [None, True, False, "text"]], False),
        ('a = [1, 2, 3, 4, 5], insert(a, len(a), {"x": null})', 'a', [1, 2, 3, 4, 5, {"x": None}], False),
        ('a = [1, 2, 3], insert(a, len(a), 10)', 'a', [1, 2, 3, 10], False),
        ('a = [1, 2, 3], insert(a, 3, 10)', 'a', [1, 2, 3, 10], False),
        ('a = [1, 2, 3], insert(a, 1, 10)', 'a', [1, 10, 2, 3], False),
        ('a = [1, 2, 3], insert(a, 0, 10)', 'a', [10, 1, 2, 3], False),
        ('a = [1, 2, 3], insert(a, -0, 10)', 'a', [10, 1, 2, 3], False),
        ('a = [1, 2, 3], insert(a, -1, 10)', 'a', [1, 2, 10, 3], False),
        ('a = [1, 2, 3], insert(a, -3, 10)', 'a', [10, 1, 2, 3], False),
        ('a = [1, 2, 3], insert(a, 4, 10)', 'a', None, True),  # Out of range
        ('a = [1, 2, 3], insert(a, -4, 10)', 'a', None, True),  # Out of range
        ('a = [1, 2, 3], insert(a, true, 10)', 'a', None, True),  # Index must be number.
        ('a = [1, 2, 3], insert(a, 3)', 'a', None, True),  # Lack the inserted value.
        ('a = [1, 2, 3], insert(a)', 'a', None, True),  # Lack index and the inserted value.
        ('a = 1, insert(a, 0, 10)', 'a', None, True),  # Array is needed.
        ('a = [1, 2, 3, 4, 5, 6], insert(a, len(a), 10)', 'a', None, True),  # Limit of max array size

        ('', 'strip("abc")', "abc", False),
        ('a = "abc"', 'strip(a)', "abc", False),
        ({'a': "\n 　\r abc \n\t"}, 'strip(a)', "abc", False),  # multi-byte character "　"
        ({'a': "\n 　\r αβ \n\t"}, 'strip(a)', "αβ", False),  # multi-byte character "　" and "αβ"
        # ('', 'strip("\n 　\r abc \n\t")', "abc", False),
        # ('a = "\n 　\r abc \n\t"', 'strip(a)', "abc", False),
        # ('a = "\n 　\r αβ \n\t"', 'strip(a)', "αβ", False),
        ('', 'strip()', None, True),  # The argument is necessary.

        ('', 'type(1)', "int", False),
        ('a = 1.5', 'type(a)', "float", False),
        ('', 'type(1.5)', "float", False),
        ('', 'type(1e3)', "float", False),
        ('', 'type("1")', "string", False),
        ('', 'type(true)', "boolean", False),
        ('', 'type(false)', "boolean", False),
        ('', 'type(null)', "null", False),
        ('', 'type([1,2,3])', "array", False),
        ('x = [1, [2, 3]]', 'type(x[1])', "array", False),
        ('', 'type([])', "array", False),
        ('', 'type({"a":1,"b":2})', "block", False),
        ('x = [1, {"a":1,"b":2}]', 'type(x[1])', "block", False),
        ('', 'type({})', "block", False),
        ('function enclosure(a) {x = a, function closure(y) {return(x + y)}, return(closure)}, z1 = enclosure(100)', 'type(z1)', "function", False),
        ('function x2(a) {return(a*2)}', 'type(x2)', "function", False),
        ('function x2(a) {return(a*2)}, foo = x2', 'type(foo)', "function", False),
        ('function x2(a) {return(a*2)}, foo = [x2]', 'type(foo.0)', "function", False),
        ('', 'type(len)', "function", False),
        ('', 'type(print)', "function", False),
        ('', 'type(for)', None, True),  # Parse error
        ('', 'type(break)', None, True),  # Parse error
        ('', 'type()', None, True),  # The argument is necessary.
        ('', 'type(1, 2)', None, True),  # Only one argument is necessary.
    )
)

def test_function(prepared_expression, target_expression, expected_value,
                  is_error):
    """Unit-test for Standard library

    Args:
        prepared_expression (str or dict): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
                                 str: Statements as VivJson
                                 dict: Value as Python's value
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_depth=10, max_array_size=6, nan='NaN',
                    enable_tag_detail=True)

    text = f'return({target_expression})'
    value, error_message = Viv.run(prepared_expression, text, config)
    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123



"""Token for VivJson

For example,

              GT  NUMBER  ASSIGN  PLUS  NUMBER
               |  |            |   |    |
    while  ( i >  3     )  { i = i +    1     }
    |      | |          |  | |   |            |
IDENTIFIER | IDENTIFIER |  | IDENTIFIER       |
           |            |  |                  |
       LEFT_PARENTHESIS |  LEFT_CURLY_BRACKET |
                        |                     |
          RIGHT_PARENTHESIS            RIGHT_CURLY_BRACKET


Refer to:
- "Let's make a Teeny Tiny compiler"
  https://austinhenley.com/blog/teenytinycompiler1.html
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB sites is not used.

Note:
This file name is "tokens.py" instead of "token.py". Because `pydoc`
does not work correctly for "token.py". Then `pydoc` outputs error
as below.
> ImportError: cannot import name 'EXACT_TOKEN_TYPES' from 'token'

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-30'

from enum import Enum

class Token:
    """Token class

    Attributes:
        type (Token.Type): The type of token
        lexeme (str or NoneType): The token's actual text
        medium (str or NoneType): The name of object where the token is
                                  existed. For example, "test.viv",
                                  "1st argument".
        line_number (int or NoneType): Line number where the token is
                                       existed.
        column_number (int or NoneType): Column number where the token
                                         is existed.
        _keywords (dict): Keywords table that is generated from
                          _KEYWORD_TYPES.
        _KEYWORD_TYPES (tuple): Token types of Keyword.
    """
    class Type(Enum):
        """Token types"""
        # 4 arithmetic operators and so on
        PLUS = '+'
        MINUS = '-'
        MULTIPLY = '*'
        DIVIDE = '/'
        MODULO = '%'

        # Specific operator
        IN = 'in'

        # Using various purpose
        DOT = '.'
        LEFT_PARENTHESIS = '('
        RIGHT_PARENTHESIS = ')'
        LEFT_SQUARE_BRACKET = '['
        RIGHT_SQUARE_BRACKET = ']'
        LEFT_CURLY_BRACKET = '{'
        RIGHT_CURLY_BRACKET = '}'

        # Assignments (Substitutes)
        ASSIGN = '='
        SET = ':'
        RESULT = ':='
        PLUS_ASSIGN = '+='
        MINUS_ASSIGN = '-='
        MULTIPLY_ASSIGN = '*='
        DIVIDE_ASSIGN = '/='
        MODULO_ASSIGN = '%='

        # Conditions
        EQ = '=='
        NE = '!='
        LT = '<'
        LE = '<='
        GT = '>'
        GE = '>='

        # Logical operators
        OR = 'or'
        AND = 'and'
        NOT = 'not'

        # Modifiers
        CLASS = 'class'
        FUNCTION = 'function'
        REFERENCE = 'reference'

        # Statements
        BREAK = 'break'
        CONTINUE = 'continue'
        RETURN = 'return'

        # Values
        NUMBER = 'number'
        STRING = 'string'
        NULL = 'null'
        TRUE = 'true'
        FALSE = 'false'

        # Variable/Function name
        IDENTIFIER = 'identifier'

        # Functions
        IF = 'if'
        ELSEIF = 'elseif'
        ELSE = 'else'
        FOR = 'for'
        REMOVE = 'remove'
        INCLUDE = 'include'
        IMPORT = 'import'
        SUPER = 'super'

        # Terminator
        NEW_LINE = 'new_line'
        SEMICOLON = ';'
        COMMA = ','
        EOS = 'eos'

        # Unexpected
        ERROR = 'error'

    _KEYWORD_TYPES = (
        Type.IN, Type.OR, Type.AND, Type.NOT,
        Type.CLASS, Type.FUNCTION, Type.REFERENCE,
        Type.BREAK, Type.CONTINUE, Type.RETURN,
        Type.NULL, Type.TRUE, Type.FALSE,
        Type.IF, Type.ELSEIF, Type.ELSE,
        Type.FOR, Type.REMOVE,
        Type.INCLUDE, Type.IMPORT, Type.SUPER
    )

    def __init__(self, token_type, medium=None,  # pylint: disable=R0913
                 line_number=None, column_number=None, lexeme=None):
        """Initialize class.

        Args:
            token_type (Token.Type): The type of token
            medium (str, optional): The name of object where the token
                                    is existed. For example,
                                    "test.viv", "1st argument".
            line_number (int, optional): Line number where the token
                                         is existed.
            column_number (int, optional): Column number where the
                                           token is existed.
            lexeme (str, optional): The token's actual text. It is so
                                    called "lexeme".
                                    When this is omitted or this is
                                    None, the assigned value of Token
                                    type is used.
        """
        self.type = token_type
        self.medium = medium
        self.line_number = line_number
        self.column_number = column_number
        self.lexeme = lexeme if lexeme is not None else token_type.value

        self._keywords = {}
        for keyword_type in self._KEYWORD_TYPES:
            self._keywords[keyword_type.value] = keyword_type

    def __repr__(self):
        return self.to_string()

    def fix_keyword_type(self):
        """Fix type if this is Keyword.

        Returns:
            bool: True if it is fixed,
                  False otherwise.
        """
        is_fixed = self.lexeme in self._keywords
        if is_fixed:
            self.type = self._keywords[self.lexeme]
        return is_fixed

    def to_string(self, omit_type_name=False):
        """Get as string.

        Args:
            omit_type_name (bool, optional): Type name is omitted when
                                             it is True.
                                             The default is False.

        Returns:
            str: Type name and lexeme
        """
        text = self.lexeme if self.lexeme != '\n' else 'LF'
        text = f'"{text}"' if omit_type_name else f'{self.type.name} ({text})'
        additional = []
        if self.medium is not None:
            additional.append(self.medium)
        if self.line_number is not None:
            additional.append(f'line: {self.line_number}')
        if self.column_number is not None:
            additional.append(f'column: {self.column_number}')
        if len(additional) > 0:
            text +=  f' in ({", ".join(additional)})'
        return text
